""" ptah-manage command """
from __future__ import print_function
import sys
import argparse
import textwrap
from collections import OrderedDict
from pyramid.compat import configparser, NativeIO

import ptah
from ptah import scripts
from ptah.manage.manage import MANAGE_ID


grpTitleWrap = textwrap.TextWrapper(
    initial_indent='* ',
    subsequent_indent='  ')

grpDescriptionWrap = textwrap.TextWrapper(
    initial_indent='    ',
    subsequent_indent='    ')


def main(init=True):
    args = ManageCommand.parser.parse_args()

    # bootstrap pyramid
    if init: # pragma: no cover
        env = scripts.bootstrap(args.config)

    cmd = ManageCommand(args)
    cmd.run()

    ptah.shutdown()


class ManageCommand(object):

    parser = argparse.ArgumentParser(description="ptah manage")
    parser.add_argument('config', metavar='config',
                        help='ini config file')
    parser.add_argument('--list-modules', action="store_true",
                        dest='modules',
                        help='List ptah management modules')

    parser.add_argument('--list-models', action="store_true",
                        dest='models',
                        help='List ptah models')

    def __init__(self, args):
        self.options = args

    def run(self):
        print ('')
        if self.options.modules:
            self.list_modules()
        elif self.options.models:
            self.list_models()
        else:
            self.parser.print_help()

    def list_modules(self):
        cfg = ptah.get_settings(ptah.CFG_ID_PTAH)
        disabled = cfg['disable_modules']

        mods = []
        for name, mod in ptah.get_cfg_storage(MANAGE_ID).items():
            mods.append(
                 {'id': name,
                  'title': mod.title,
                  'description': mod.__doc__,
                  'disabled': name in disabled})

        for mod in sorted(mods, key=lambda item:item['id']):
            print (grpTitleWrap.fill(
                    '{id}: {title} (disabled: {disabled})'.format(**mod)))
            print (grpDescriptionWrap.fill(mod['description']))
            print ('\n')

    def list_models(self):
        cfg = ptah.get_settings(ptah.CFG_ID_PTAH)
        disabled = cfg['disable_models']

        types = []
        for ti in ptah.cms.get_types().values():
            types.append(
                {'name': ti.__uri__,
                 'title': ti.title,
                 'description': ti.description,
                 'disabled': ti.__uri__ in disabled,
                 'cls': ti.cls})

        for ti in sorted(types, key=lambda item:item['name']):
            print (grpTitleWrap.fill(
                    '{name}: {title} (disabled: {disabled})'.format(**ti)))
            if ti['description']:
                print (grpDescriptionWrap.fill(ti['description']))
            print('')

            cls = ti['cls']
            print(grpDescriptionWrap.fill('class: {0}'.format(cls.__name__)))
            print(grpDescriptionWrap.fill('module: {0}'.format(cls.__module__)))
            print('    file: ', sys.modules[cls.__module__].__file__)
            print('\n')



import datetime
from ptah.form import iso8601


def test_utc():
    dt = datetime.datetime.now()
    assert iso8601.UTC.tzname(dt) == 'UTC'
    assert iso8601.UTC.dst(dt) == datetime.timedelta(0)

def test_iso8601_regex():
    assert iso8601.ISO8601_REGEX.match("2006-10-11T00:14:33Z")

def test_timezone_regex():
    assert iso8601.TIMEZONE_REGEX.match("+01:00")
    assert iso8601.TIMEZONE_REGEX.match("+00:00")
    assert iso8601.TIMEZONE_REGEX.match("+01:20")
    assert iso8601.TIMEZONE_REGEX.match("-01:00")

def test_parse_date():
    d = iso8601.parse_date("2006-10-20T15:34:56Z")
    assert d.year == 2006
    assert d.month == 10
    assert d.day == 20
    assert d.hour == 15
    assert d.minute == 34
    assert d.second == 56
    assert d.tzinfo == iso8601.UTC

def test_parse_date_fraction():
    d = iso8601.parse_date("2006-10-20T15:34:56.123Z")
    assert d.year == 2006
    assert d.month == 10
    assert d.day == 20
    assert d.hour == 15
    assert d.minute == 34
    assert d.second == 56
    assert d.microsecond == 123000
    assert d.tzinfo == iso8601.UTC

def test_parse_date_fraction_2():
    """From bug 6

    """
    d = iso8601.parse_date("2007-5-7T11:43:55.328Z'")
    assert d.year == 2007
    assert d.month == 5
    assert d.day == 7
    assert d.hour == 11
    assert d.minute == 43
    assert d.second == 55
    assert d.microsecond == 328000
    assert d.tzinfo == iso8601.UTC

def test_parse_date_tz():
    d = iso8601.parse_date("2006-10-20T15:34:56.123+02:30")
    assert d.year == 2006
    assert d.month == 10
    assert d.day == 20
    assert d.hour == 15
    assert d.minute == 34
    assert d.second == 56
    assert d.microsecond == 123000
    assert d.tzinfo.tzname(None) == "+02:30"
    offset = d.tzinfo.utcoffset(None)
    assert offset.days == 0
    assert offset.seconds == 60 * 60 * 2.5

def test_parse_invalid_date():
    try:
        iso8601.parse_date(None)
    except iso8601.ParseError:
        pass
    else: # pragma: no cover
        assert 1 == 2

def test_parse_invalid_date2():
    try:
        iso8601.parse_date("23")
    except iso8601.ParseError:
        pass
    else: # pragma: no cover
        assert 1 == 2

def test_parse_no_timezone():
    """issue 4 - Handle datetime string without timezone

    This tests what happens when you parse a date with no timezone. While not
    strictly correct this is quite common. I'll assume UTC for the time zone
    in this case.
    """
    d = iso8601.parse_date("2007-01-01T08:00:00")
    assert d.year == 2007
    assert d.month == 1
    assert d.day == 1
    assert d.hour == 8
    assert d.minute == 0
    assert d.second == 0
    assert d.microsecond == 0
    assert d.tzinfo == iso8601.UTC

def test_parse_no_timezone_different_default():
    tz = iso8601.FixedOffset(2, 0, "test offset")
    d = iso8601.parse_date("2007-01-01T08:00:00", default_timezone=tz)
    assert d.tzinfo == tz

def test_space_separator():
    """Handle a separator other than T

    """
    d = iso8601.parse_date("2007-06-23 06:40:34.00Z")
    assert d.year == 2007
    assert d.month == 6
    assert d.day == 23
    assert d.hour == 6
    assert d.minute == 40
    assert d.second == 34
    assert d.microsecond == 0
    assert d.tzinfo == iso8601.UTC



from ptah.testing import PtahTestCase, TestCase
from webob.multidict import MultiDict
from pyramid.compat import text_
from pyramid.view import view_config, render_view_to_response
from pyramid.testing import DummyRequest
from pyramid.httpexceptions import HTTPFound, HTTPForbidden


class TestFormWidgets(TestCase):
    def _makeOne(self, fields, form, request):
        from ptah.form.form import FormWidgets
        return FormWidgets(fields, form, request)

    def test_ctor(self):
        request = DummyRequest()
        form = DummyForm()
        fields = DummyFields()
        inst = self._makeOne(fields, form, request)
        self.assertEqual(inst.form_fields, fields)
        self.assertEqual(inst.form, form)
        self.assertEqual(inst.request, request)


class TestFormErrors(PtahTestCase):

    def test_form_errors(self):
        import ptah
        from ptah.form import Invalid, TextField
        from ptah.form.form import formErrorMessage
        request = DummyRequest()

        err1 = Invalid(None, 'Error1')
        err2 = Invalid(None, 'Error2')
        err2.field = TextField('text')

        msg = ptah.view.Message([err1, err2], request)

        errs = formErrorMessage(msg, request)['errors']
        self.assertIn(err1, errs)
        self.assertNotIn(err2, errs)

        ptah.view.add_message(request, [err1, err2], 'form-error')

        res = ptah.view.render_messages(request)
        self.assertIn('Please fix indicated errors.', res)


class TestForm(PtahTestCase):

    def test_basics(self):
        from ptah.form.form import Form

        request = DummyRequest()
        form = Form(None, request)

        request.url = '/test/form'
        self.assertEqual(form.action, request.url)

        self.assertEqual(form.name, 'form')

        form = Form(None, request)
        form.prefix = 'my.test.form.'
        self.assertEqual(form.name, 'my.test.form')

        self.assertEqual(form.id, 'my-test-form')

    def test_form_content(self):
        from ptah.form.form import Form

        request = DummyRequest()
        form = Form(None, request)

        self.assertIsNone(form.form_content())

        form_content = {}
        form.content = form_content
        self.assertIs(form.form_content(), form_content)

    def test_form_content_from_update(self):
        from ptah.form.form import Form

        request = DummyRequest()
        form = Form(None, request)

        form_content = {'test': 'test1'}
        form.update(**form_content)
        self.assertEqual(form.form_content(), form_content)

    def test_csrf_token(self):
        from ptah.form import form

        class MyForm(form.Form):
            pass

        request = DummyRequest()
        form_ob = MyForm(None, request)

        token = form_ob.token
        self.assertEqual(token, request.session.get_csrf_token())
        self.assertIsNotNone(token)
        self.assertIsNone(form_ob.validate_csrf_token())

        request.POST = {}

        form_ob.csrf = True
        self.assertRaises(HTTPForbidden, form_ob.validate_csrf_token)
        self.assertRaises(HTTPForbidden, form_ob.validate, {}, [])

        request.POST = {form_ob.csrfname: token}
        self.assertIsNone(form_ob.validate_csrf_token())

    def test_form_params(self):
        from ptah.form.form import Form, DisplayForm

        request = DummyRequest()

        form = Form(None, request)
        disp_form = DisplayForm(None, request)

        self.assertEqual(form.method, 'post')

        post = {'post': 'info'}
        request.POST = post

        self.assertIs(form.form_params(), post)
        self.assertIs(disp_form.form_params(), DisplayForm.params)

        get = {'get': 'info'}
        request.GET = get
        form.method = 'get'
        self.assertIs(form.form_params(), get)

        form.method = 'unknown'
        self.assertEqual(form.form_params(), None)

        params = {'params': 'info'}
        form.method = 'POST'
        form.params = params
        self.assertIn('params', form.form_params().keys())
        self.assertIsInstance(form.form_params(), MultiDict)

        params = MultiDict({'params': 'info'})
        form.method = 'POST'
        form.params = params
        self.assertIs(form.form_params(), params)

    def test_form_params_method(self):
        from ptah.form.form import Form

        form = Form(None, None)
        form.method = 'params'
        params = {'post': 'info'}
        form.params = params

        self.assertEqual(list(form.form_params().keys()), ['post'])
        self.assertEqual(list(form.form_params().values()), ['info'])

    def test_form_mode(self):
        from ptah.form.form import Form, DisplayForm, \
            FORM_INPUT, FORM_DISPLAY

        request = DummyRequest()

        form = Form(None, request)
        self.assertEqual(form.mode, FORM_INPUT)

        form = DisplayForm(None, request)
        self.assertEqual(form.mode, FORM_DISPLAY)

    def test_form_update_widgets(self):
        from ptah import form

        request = DummyRequest()
        request.POST = {}

        form_ob = form.Form(None, request)
        form_ob.update()

        self.assertIsInstance(form_ob.widgets, form.FormWidgets)
        self.assertEqual(form_ob.widgets.mode, form_ob.mode)

        form_ob.mode = form.FORM_DISPLAY
        form_ob.update()
        self.assertEqual(form_ob.widgets.mode, form.FORM_DISPLAY)

        self.assertEqual(len(form_ob.widgets), 0)

        form_ob.fields = form.Fieldset(form.TextField('test'))
        form_ob.update()
        self.assertEqual(len(form_ob.widgets), 1)
        self.assertIn('test', form_ob.widgets)
        self.assertIn('test', [f.name for f in form_ob.widgets.fields()])

        self.assertIsInstance(form_ob.widgets['test'], form.TextField)
        self.assertEqual(form_ob.widgets['test'].name, 'test')
        self.assertEqual(form_ob.widgets['test'].id, 'form-widgets-test')

    def test_form_extract(self):
        from ptah import form

        request = DummyRequest()
        request.POST = {}

        form_ob = form.Form(None, request)
        form_ob.fields = form.Fieldset(form.TextField('test'))
        form_ob.update()

        data, errors = form_ob.extract()
        self.assertEqual(errors[0].msg, 'Required')

        request.POST = {'test': 'Test string'}
        form_ob.update()
        data, errors = form_ob.extract()
        self.assertEqual(data['test'], 'Test string')

    def test_form_render(self):
        from ptah import form

        request = DummyRequest()

        form_ob = form.Form(None, request)
        form_ob.fields = form.Fieldset(form.TextField('test'))
        form_ob.update()

        self.assertIn('<form action="http://example.com"', form_ob.render())

    def test_form_render_view_config_renderer(self):
        from ptah import form
        request = DummyRequest()

        class CustomForm(form.Form):
            fields = form.Fieldset(form.TextField('test'))

        self.config.add_view(
            name='test', view=CustomForm,
            renderer='ptah.form:tests/test-form.pt')

        resp = render_view_to_response(None, request, 'test', False).body

        self.assertIn(b'<h1>Custom form</h1>', resp)
        self.assertIn(b'<form action="http://example.com"', resp)

    def test_form_render_view_config(self):
        from ptah import form
        request = DummyRequest()

        class CustomForm(form.Form):
            fields = form.Fieldset(form.TextField('test'))

        self.config.add_view(name='test', view=CustomForm)

        resp = render_view_to_response(None, request, 'test', False).body
        self.assertIn('<form action="http://example.com"', text_(resp))

    def test_form_render_view_config_return(self):
        from ptah import form
        request = DummyRequest(POST={'form.buttons.test': 'test'})

        class CustomForm(form.Form):
            fields = form.Fieldset(form.TextField('test'))

            @form.button('test')
            def handler(self):
                return HTTPFound(location='.')

        res = CustomForm(object(), request)()
        self.assertIsInstance(res, HTTPFound)


class DummyForm(object):
    context = None
    prefix = 'prefix'
    def form_params(self): # pragma: no cover
        return None
    def form_content(self): # pragma: no cover
        return None


class DummyFieldset(object):
    def fieldsets(self): # pragma: no cover
        return []


class DummyFields(object):
    def __init__(self, fieldset=None):
        if fieldset is None:
            fieldset = DummyFieldset()
        self.fieldset = fieldset

    def bind(self, data, params, context=None): # pragma: no cover
        self.data = data
        self.params = params
        self.context = context
        return self.fieldset



""" populate data """
import logging
import transaction
from pyramid.request import Request
from pyramid.interfaces import IRequestFactory
from pyramid.threadlocal import manager as threadlocal_manager

import ptah
from ptah import config
from ptah.migrate import update_versions

POPULATE_ID = 'ptah:populate-step'
POPULATE_DB_SCHEMA = 'ptah-db-schema'


class populate(object):
    """Registers a data populate step. Populate steps are used by
    :ref:`data_populate_script` command line tool and by
    :ref:`ptah_populate_dir` pyramid directive for populate system data.

    :param name: Unique step name
    :param title: Human readable title
    :param active: Should this step automaticly executed or not
    :param requires: List of steps that should be executed before this step

    Populate step interface :py:class:`ptah.interfaces.populate_step`.
    Steps are executed after configuration is completed.

    .. code-block:: python

       import ptah

       @ptah.populate('custom-user',
                      title='Create custom user',
                      requires=(ptah.POPULATE_DB_SCHEMA,))
       def create_custom_user(registry):
           # create user

    ``create_custom_user`` executes only after ``ptah.POPULATE_DB_SCHEMA`` step.

    Perpose of inactive steps is for example entering testing data or executing
    custom step.
    """

    def __init__(self, name, title='', active=True, requires=(), __depth=1):
        self.info = config.DirectiveInfo(__depth)
        self.discr = (POPULATE_ID, name)

        self.intr = intr = config.Introspectable(
            POPULATE_ID, self.discr, name, POPULATE_ID)

        intr['name'] = name
        intr['title'] = title
        intr['active'] = active
        intr['requires'] = requires
        intr['codeinfo'] = self.info.codeinfo

    @classmethod
    def pyramid(cls, cfg, name, factory=None,
                title='', active=True, requires=()):
        """ Pyramid `ptah_populate_step` directive:

        .. code-block:: python

          config = Configurator()
          config.include('ptah')

          config.ptah_populate_step('ptah-create-db-schema', factory=..)
        """
        l = populate(name, title, active, requires, 3)(factory, cfg)

    def __call__(self, factory, cfg=None):
        intr = self.intr
        intr['factory'] = factory

        self.info.attach(
            config.Action(
                lambda cfg, name, intr:
                    cfg.get_cfg_storage(POPULATE_ID).update({name: intr}),
                (intr['name'], intr),
                   discriminator=self.discr, introspectables=(intr,)),
            cfg)
        return factory


class Populate(object):

    def __init__(self, registry):
        self.registry = registry

    def list_steps(self, p_steps=None, all=False):
        seen = set()

        steps = dict(
            (name, intr) for name, intr in
            ptah.get_cfg_storage(POPULATE_ID, self.registry).items())

        sorted_steps = []
        def _step(name, step):
            if name in seen:
                return

            seen.add(name)

            for dep in step['requires']:
                if dep not in steps:
                    raise RuntimeError(
                        "Can't find populate step '{0}'.".format(dep))
                _step(dep, steps[dep])

            sorted_steps.append(step)

        if p_steps is not None:
            for name in p_steps:
                if name not in steps:
                    raise RuntimeError(
                        "Can't find populate step '{0}'.".format(name))

                _step(name, steps[name])
        else:
            for name, step in steps.items():
                if all:
                    _step(name, step)
                elif step['active']:
                    _step(name, step)

        return sorted_steps

    def execute(self, p_steps=None, request=None):
        registry = self.registry
        if request is None:
            request_factory = registry.queryUtility(
                IRequestFactory, default=Request)
            request = request_factory.blank('/')
            request.registry = registry

        threadlocals = {'registry':registry, 'request':request}
        threadlocal_manager.push(threadlocals)

        steps = self.list_steps(p_steps)

        log = logging.getLogger(' ptah ')

        for step in steps:
            log.info('Executing populate step: %s', step['name'])
            step['factory'](registry)

        transaction.commit()
        threadlocal_manager.pop()


@populate(POPULATE_DB_SCHEMA, title='Create db schema')
def create_db_schema(registry, update=True):
    registry.notify(ptah.events.BeforeCreateDbSchema(registry))

    skip_tables = ptah.get_settings(ptah.CFG_ID_PTAH)['db_skip_tables']

    Base = ptah.get_base()

    log = logging.getLogger('ptah')

    tables = []
    for name, table in Base.metadata.tables.items():
        if name not in skip_tables and not table.exists():
            log.info("Creating db table `%s`.", name)
            tables.append(table)

    if tables:
        Base.metadata.create_all(tables=tables)
        transaction.commit()

    # update db versions
    if update:
        update_versions(registry)



""" ptah settings """
import pytz
import logging
import sqlalchemy
import translationstring
from email.utils import formataddr

from pyramid.compat import bytes_
from pyramid.events import ApplicationCreated

import ptah
from ptah import manage, settings

_ = translationstring.TranslationStringFactory('ptah')

log = logging.getLogger('ptah')


ptah.register_settings(
    ptah.CFG_ID_PTAH,

    ptah.form.BoolField(
        'auth',
        title = _('Authentication policy'),
        description = _('Enable authentication policy.'),
        default = False),

    ptah.form.TextField(
        'secret',
        title = _('Authentication policy secret'),
        description = _('The secret (a string) used for auth_tkt '
                        'cookie signing'),
        default = '',
        tint = True),

    ptah.form.TextField(
        'manage',
        title = 'Ptah manage id',
        default = 'ptah-manage'),

    ptah.form.LinesField(
        'managers',
        title = 'Manage login',
        description = 'List of user logins with access rights to '\
                            'ptah management ui.',
        default = ()),

    ptah.form.TextField(
        'manager_role',
        title = 'Manager role',
        description = 'Specific role with access rights to ptah management ui.',
        default = ''),

    ptah.form.LinesField(
        'disable_modules',
        title = 'Hide Modules in Management UI',
        description = 'List of modules names to hide in manage ui',
        default = ()),

    ptah.form.LinesField(
        'enable_modules',
        title = 'Enable Modules in Management UI',
        description = 'List of modules names to enable in manage ui',
        default = ()),

    ptah.form.LinesField(
        'disable_models',
        title = 'Hide Models in Model Management UI',
        description = 'List of models to hide in model manage ui',
        default = ()),

    ptah.form.TextField(
        'email_from_name',
        default = 'Site administrator'),

    ptah.form.TextField(
        'email_from_address',
        validator = ptah.form.Email(),
        required = False,
        default = 'admin@localhost'),

    ptah.form.ChoiceField(
        'pwd_manager',
        title = 'Password manager',
        description = 'Available password managers '\
            '("plain", "ssha", "bcrypt")',
        vocabulary = ptah.form.SimpleVocabulary.from_values(
            "plain", "ssha",),
        default = 'plain'),

    ptah.form.IntegerField(
        'pwd_min_length',
        title = 'Length',
        description = 'Password minimium length.',
        default = 5),

    ptah.form.BoolField(
        'pwd_letters_digits',
        title = 'Letters and digits',
        description = 'Use letters and digits in password.',
        default = False),

    ptah.form.BoolField(
        'pwd_letters_mixed_case',
        title = 'Letters mixed case',
        description = 'Use letters in mixed case.',
        default = False),

    ptah.form.LinesField(
        'db_skip_tables',
        title = 'Skip table creation',
        description = 'Do not create listed tables during data population.',
        default = ()),

    ptah.form.LinesField(
        'default_roles',
        title = 'Default roles',
        description = 'List of default assigned roles for all principals.',
        default = ()),

    title = _('Ptah settings'),
)


ptah.register_settings(
    ptah.CFG_ID_FORMAT,

    ptah.form.TimezoneField(
        'timezone',
        default = pytz.timezone('US/Central'),
        title = _('Timezone'),
        description = _('Site wide timezone.')),

    ptah.form.TextField(
        'date_short',
        default = '%m/%d/%y',
        title = _('Date'),
        description = _('Date short format')),

    ptah.form.TextField(
        'date_medium',
        default = '%b %d, %Y',
        title = _('Date'),
        description = _('Date medium format')),

    ptah.form.TextField(
        'date_long',
        default = '%B %d, %Y',
        title = _('Date'),
        description = _('Date long format')),

    ptah.form.TextField(
        'date_full',
        default = '%A, %B %d, %Y',
        title = _('Date'),
        description = _('Date full format')),

    ptah.form.TextField(
        'time_short',
        default = '%I:%M %p',
        title = _('Time'),
        description = _('Time short format')),

    ptah.form.TextField(
        'time_medium',
        default = '%I:%M %p',
        title = _('Time'),
        description = _('Time medium format')),

    ptah.form.TextField(
        'time_long',
        default = '%I:%M %p %z',
        title = _('Time'),
        description = _('Time long format')),

    ptah.form.TextField(
        'time_full',
        default = '%I:%M:%S %p %Z',
        title = _('Time'),
        description = _('Time full format')),

    title = 'Site formats',
    )


def set_mailer(cfg, mailer):
    def action(cfg, mailer):
        PTAH = ptah.get_settings(ptah.CFG_ID_PTAH, cfg.registry)
        PTAH['Mailer'] = mailer

    cfg.action('ptah.ptah_mailer', action, (cfg, mailer))


class DummyMailer(object):

    def send(self, from_, to_, message):
        log.warning("Mailer is not configured.")
        log.warning(message)


@ptah.subscriber(ptah.events.SettingsInitializing)
def initialized(ev):
    PTAH = ptah.get_settings(ptah.CFG_ID_PTAH, ev.registry)

    # mail
    if PTAH.get('Mailer') is None:
        PTAH['Mailer'] = DummyMailer()
        PTAH['full_email_address'] = formataddr(
            (PTAH['email_from_name'], PTAH['email_from_address']))


def enable_manage(cfg, name='ptah-manage', access_manager=None,
                  managers=None, manager_role=None,
                  disable_modules=None, enable_modules=None):
    """Implementation for pyramid `ptah_init_manage` directive """
    def action(cfg, name, access_manager,
               managers, manager_role, disable_modules):
        PTAH = cfg.ptah_get_settings(ptah.CFG_ID_PTAH)

        PTAH['manage'] = name
        if managers is not None:
            PTAH['managers'] = managers
        if manager_role is not None:
            PTAH['manager_role'] = manager_role
        if disable_modules is not None:
            PTAH['disable_modules'] = disable_modules
        if enable_modules is not None:
            PTAH['enable_modules'] = enable_modules

        if access_manager is None:
            access_manager = ptah.manage.PtahAccessManager(cfg.registry)

        ptah.manage.set_access_manager(access_manager, cfg.registry)

    cfg.add_route('ptah-manage', '/{0}/*traverse'.format(name),
                  factory=ptah.manage.PtahManageRoute, use_global_views=True)
    cfg.action(
        'ptah.ptah_manage', action,
        (cfg, name, access_manager,
         managers, manager_role, disable_modules), order=999999+1)


def initialize_sql(cfg, prefix='sqlalchemy.'):
    def action(cfg, cache):
        PTAH = cfg.ptah_get_settings(ptah.CFG_ID_PTAH)
        PTAH['sqlalchemy_cache'] = {}
        PTAH['sqlalchemy_initialized'] = True

    cache = {}
    engine = sqlalchemy.engine_from_config(
        cfg.registry.settings, prefix,
        execution_options = {'compiled_cache': cache, 'echo': True})

    ptah.get_session().configure(bind=engine)
    ptah.get_session_maker().configure(bind=engine)
    ptah.get_base().metadata.bind = engine

    cfg.action('ptah.initalize_sql', action, (cfg, cache))

    # check_version
    from ptah import migrate
    cfg.add_subscriber(migrate.check_version, ApplicationCreated)


@ptah.subscriber(ApplicationCreated)
def starting(ev):
    settings.load_dbsettings()



import ptah
from ptah.testing import PtahTestCase
from pyramid.testing import DummyRequest
from pyramid.view import render_view_to_response


class TestPermissionsModule(PtahTestCase):

    def test_perms_module(self):
        from ptah.manage.manage import PtahManageRoute
        from ptah.manage.permissions import PermissionsModule

        request = DummyRequest()

        ptah.auth_service.set_userid('test')
        cfg = ptah.get_settings(ptah.CFG_ID_PTAH, self.registry)
        cfg['managers'] = ('*',)
        mr = PtahManageRoute(request)
        mod = mr['permissions']

        self.assertIsInstance(mod, PermissionsModule)

    def test_perms_view(self):
        from ptah.manage.permissions import PermissionsModule, PermissionsView

        request = DummyRequest()

        mod = PermissionsModule(None, request)

        res = render_view_to_response(mod, request, '', False)
        self.assertEqual(res.status, '200 OK')

    def test_perms_roles(self):
        from ptah.manage.permissions import PermissionsModule, RolesView

        request = DummyRequest()

        mod = PermissionsModule(None, request)

        res = render_view_to_response(mod, request, 'roles.html', False)
        self.assertEqual(res.status, '200 OK')
        self.assertIn('<h2>Roles</h2>', res.text)



""" base class """
import sys
import sqlalchemy
import transaction
import pkg_resources
from zope.interface import directlyProvides
from pyramid import testing
from pyramid.interfaces import \
     IRequest, IAuthenticationPolicy, IAuthorizationPolicy
from pyramid.interfaces import IRouteRequest
from pyramid.view import render_view, render_view_to_response
from pyramid.path import package_name
from pyramid.authorization import ACLAuthorizationPolicy
from pyramid.authentication import AuthTktAuthenticationPolicy

if sys.version_info[:2] == (2, 6): # pragma: no cover
    from unittest2 import TestCase
else:
    from unittest import TestCase

import ptah
from ptah import config


class PtahTestCase(TestCase):

    _init_ptah = True
    _init_sqla = True

    _settings = {'sqlalchemy.url': 'sqlite://'}
    _packages = ()
    _trusted_manage = True
    _environ = {
        'wsgi.url_scheme':'http',
        'wsgi.version':(1,0),
        'HTTP_HOST': 'example.com',
        'SCRIPT_NAME': '',
        'PATH_INFO': '/',}

    def make_request(self, environ=None, request_iface=IRequest, **kwargs):
        if environ is None:
            environ=self._environ
        request = testing.DummyRequest(environ=environ, **kwargs)
        request.request_iface = IRequest
        return request

    def init_ptah(self, *args, **kw):
        self.registry.settings.update(self._settings)
        self.config.include('ptah')

        pkg = package_name(sys.modules[self.__class__.__module__])
        if pkg != 'ptah':
            packages = []
            parts = self.__class__.__module__.split('.')
            for l in range(len(parts)):
                pkg = '.'.join(parts[:l+1])
                if pkg == 'ptah' or pkg.startswith('ptah.'):
                    continue # pragma: no cover
                try:
                    self.config.include(pkg)
                except: # pragma: no cover
                    pass

        self.config.scan(self.__class__.__module__)

        self.config.commit()
        self.config.autocommit = True

        self.config.ptah_init_settings()

        ptah.reset_session()

        if self._init_sqla:
            # create engine
            self.config.ptah_init_sql()

            # create sql tables
            Base = ptah.get_base()
            Base.metadata.create_all()
            transaction.commit()

        if self._trusted_manage:
            def trusted(*args):
                return True
            ptah.manage.set_access_manager(trusted)

    def init_pyramid(self):
        self.request = request = self.make_request()
        self.config = testing.setUp(
            request=request, settings=self._settings, autocommit=False)
        self.config.get_routes_mapper()
        self.registry = self.config.registry
        self.request.registry = self.registry

    def setUp(self):
        ptah.QueryFreezer._testing = True
        self.init_pyramid()

        if self._init_ptah:
            self.init_ptah()

    def tearDown(self):
        import ptah.util
        ptah.util.tldata.clear()

        import ptah.security
        ptah.security.DEFAULT_ACL[:] = []

        from ptah.config import ATTACH_ATTR

        mod = sys.modules[self.__class__.__module__]
        if hasattr(mod, ATTACH_ATTR):
            delattr(mod, ATTACH_ATTR)

        testing.tearDown()
        transaction.abort()

    def render_route_view(self, context, request, route_name, view=''): # pragma: no cover
        directlyProvides(
            request, self.registry.getUtility(IRouteRequest, route_name))

        return render_view_to_response(context, request, view)



""" various fields """
import datetime

import ptah
from ptah import view, formatter
from ptah.form.interfaces import _, null, Invalid
from ptah.form.field import field
from ptah.form.fields import TextAreaField, TextField, DateTimeField


@field('tinymce')
@field('ckeditor')
class CKEditorField(TextAreaField):
    """ CKEditor input widget. Field name is ``ckeditor``.

    Extra params:

    :param width: Width of widget, default is ``400px``
    :param height: Height os widget, default is ``300px``
    """

    klass = 'ckeditor-widget'

    width = '400px'
    height = '300px'

    tmpl_input = "ptah.form:templates/fields/ckeditor-input.pt"


@field('date')
class JSDateField(TextField):
    """Date input widget with JQuery Datepicker. Field name is ``date``."""

    klass = 'date-widget'
    value = ''

    tmpl_input = "ptah.form:templates/fields/jsdate-input.pt"

    def serialize(self, value):
        if value is null or value is None:
            return null

        if isinstance(value, datetime.datetime):
            value = value.date()

        if not isinstance(value, datetime.date):
            raise Invalid(self,
                          _('"${val}" is not a date object',
                            mapping={'val': value}))

        return value.strftime('%m/%d/%Y')

    def deserialize(self, value):
        if not value:
            return null
        try:
            return datetime.datetime.strptime(value, '%m/%d/%Y').date()
        except Exception as e:
            raise Invalid(
                self, _('Invalid date', mapping={'val': value, 'err': e}))


@field('datetime')
class JSDateTimeField(DateTimeField):
    """DateTime input widget with JQuery Datepicker.
    Field name is ``datetime``."""

    klass = 'datetime-widget'
    value = ''

    time_part = null
    date_part = null
    tzinfo = None

    tmpl_input = "ptah.form:templates/fields/jsdatetime-input.pt"

    def update(self, request):
        self.date_name = '%s.date' % self.name
        self.time_name = '%s.time' % self.name

        super(JSDateTimeField, self).update(request)

        self.date_part = self.params.get(self.date_name, null)
        self.time_part = self.params.get(self.time_name, null)

        if self.value:
            raw = self.value
            self.tzinfo = raw.tzinfo
            if self.date_part is null:
                self.date_part = raw.strftime('%m/%d/%Y')
            if self.time_part is null:
                FORMAT = ptah.get_settings(ptah.CFG_ID_FORMAT, request.registry)
                self.time_part = raw.strftime(FORMAT['time_short'])

        if self.date_part is null:
            self.date_part = ''
        if self.time_part is null:
            self.time_part = ''

    def extract(self, default=null):
        date = self.params.get(self.date_name, default)
        if date is default:
            return default

        if not date:
            return null

        time = self.params.get(self.time_name, default)
        if time is default:
            return default

        if not time:
            return null

        FORMAT = ptah.get_settings(ptah.CFG_ID_FORMAT, self.request.registry)
        format = '%s %s' % ('%m/%d/%Y', FORMAT['time_short'])
        try:
            dt = datetime.datetime.strptime('%s %s' % (date, time), format)
        except ValueError:
            return null

        return dt.replace(tzinfo=self.tzinfo).isoformat()



import unittest
from pyramid.registry import Registry


class Test_bootstrap(unittest.TestCase):

    def setUp(self):
        import ptah.scripts
        import pyramid.paster
        self.original_get_app = ptah.scripts.get_app
        self.original_global_registries = ptah.scripts.global_registries

        self.app = app = object()
        self.registry = registry = Registry()

        class DummyGetApp(object):

            last = registry

            def __call__(self, *a, **kw):
                self.a = a
                self.kw = kw
                return app

        pyramid.paster.get_app = DummyGetApp()
        ptah.scripts.global_registries = DummyGetApp()

    def tearDown(self):
        import ptah.scripts
        import pyramid.paster
        pyramid.paster.get_app = self.original_get_app
        ptah.scripts.global_registries = self.original_global_registries

    def test_boostrap(self):
        import ptah
        from ptah.scripts import bootstrap

        result = bootstrap('/foo/settings.ini')

        self.assertEqual(result['app'], self.app)
        self.assertEqual(result['registry'], self.registry)
        self.assertTrue(result['request'] is not None)
        self.assertEqual(result['request'].registry, self.registry)
        self.assertFalse(ptah.POPULATE)



""" interfaces """
import translationstring
from zope import interface

_ = translationstring.TranslationStringFactory('ptah')


def resolver(uri):
    """Resolve uri to object.

    :param uri: Uri string
    :rtype: Resolved object
    """


class Principal(object):
    """ Principal interface

    .. attribute:: __uri__

       Principal uri

    .. attribute:: name

       Principal name

    .. attribute:: login

       Principal login

    """


class AuthInfo(object):
    """ Authentication information

    .. attribute:: __uri__

       Principal uri or None if principal is not set

    .. attribute:: principal

       :py:class:`ptah.interfaces.Principal` object

    .. attribute:: status

       Status, True is principal has been authenticated, false otherwise

    .. attribute:: message

       Readable message from auth checkers

    """


def auth_checker(info):
    """ Perform additional checks on principal during authentication process.

    :param info: A instance of :py:class:`ptah.interfaces.AuthInfo` class
    """


class AuthProvider(object):
    """ Authentication provider interface """

    def authenticate(self, credentials):
        """ Authenticate credentials,
        return :py:class:`ptah.interfaces.Principal` object or None """

    def get_principal_bylogin(self, login):
        """ return instance of :py:class:`ptah.interfaces.Principal` of None """


def principal_searcher(term):
    """ Search users by term

    :param term: search term (str)
    :rtype: iterator of :py:class:`ptah.interfaces.Principal` instances

    """

class IOwnersAware(interface.Interface):
    """ Owners aware context

    .. attribute:: __owner__

       Owner principal uri
    """

    __owner__ = interface.Attribute('Owner')


class ILocalRolesAware(interface.Interface):
    """ Local roles aware context

    .. attribute:: __local_roles__

    """

    __local_roles__ = interface.Attribute('Local roles dict')


class IACLsAware(interface.Interface):
    """ acl maps aware context """

    __acls__ = interface.Attribute('List of acl map ids')


def populate_step(registry):
    """ Populate data step.

    :param registry: Pyramid :py:class:`pyramid.registry.Registry` object
    """


def roles_provider(context, uid, registry):
    """ Roles provider interface

    :param context: Current context object
    :param userid: User id
    :param registry: Pyramid registry object
    :rtype: Sequence of roles
    """



import sqlalchemy as sqla
import transaction
import ptah
from ptah import cms
from ptah.testing import PtahTestCase
from webob.multidict import MultiDict
from pyramid.testing import DummyRequest
from pyramid.httpexceptions import HTTPFound
from pyramid.view import render_view_to_response


class Base(PtahTestCase):

    def setUp(self):
        global Content1, Content2
        class Content1(cms.Content):
            __type__ = cms.Type('content1')

        class Content2(cms.Node):
            __tablename__ = 'test_model_base'
            __table_args__ = {'extend_existing': True}
            __id__ = sqla.Column(
                'id', sqla.Integer,
                sqla.ForeignKey('ptah_nodes.id'), primary_key=True)

            __type__ = cms.Type('content2')

            title = sqla.Column(sqla.Unicode, default='')

        self.Content1 = Content1
        self.Content2 = Content2

        super(Base, self).setUp()


class TestModelModule(Base):

    def test_model_module(self):
        from ptah.manage.model import ModelModule
        from ptah.manage.manage import PtahManageRoute

        request = DummyRequest()

        ptah.auth_service.set_userid('test')
        cfg = ptah.get_settings(ptah.CFG_ID_PTAH, self.registry)
        cfg['managers'] = ('*',)
        mr = PtahManageRoute(request)
        mod = mr['models']

        self.assertIsInstance(mod, ModelModule)

    def test_model_view(self):
        from ptah.manage.model import ModelModule, ModelModuleView

        request = DummyRequest()

        mod = ModelModule(None, request)

        res = render_view_to_response(mod, request, '', False)
        self.assertEqual(res.status, '200 OK')
        self.assertIn('content1', res.text)
        self.assertIn('content2', res.text)

    def test_model_view_disabled(self):
        from ptah.manage.model import ModelModule, ModelModuleView

        cfg = ptah.get_settings(ptah.CFG_ID_PTAH, self.registry)
        cfg['disable_models'] = ['cms-type:content2']

        request = DummyRequest()

        mod = ModelModule(None, request)

        res = render_view_to_response(mod, request, '', False)
        self.assertEqual(res.status, '200 OK')
        self.assertIn('content1', res.text)
        self.assertNotIn('content2', res.text)

    def test_model_traverse(self):
        from ptah.manage.model import ModelModule, Model

        request = DummyRequest()

        mod = ModelModule(None, request)

        item = mod['content1']
        self.assertIsInstance(item, Model)
        self.assertIs(item.tinfo, Content1.__type__)

        item = mod['content2']
        self.assertIsInstance(item, Model)
        self.assertIs(item.tinfo, Content2.__type__)

        self.assertRaises(KeyError, mod.__getitem__, 'unknown')


class TestModel(Base):

    def test_model_traverse(self):
        from ptah.manage.model import ModelModule, Record

        content = Content1()
        content.title = 'Content test'

        Session = ptah.get_session()
        Session.add(content)
        Session.flush()

        rowid = content.__id__
        transaction.commit()

        mod = ModelModule(None, DummyRequest())
        model = mod['content1']

        rec = model[str(rowid)]
        self.assertIsNotNone(rec)
        self.assertIsInstance(rec, Record)
        self.assertIsInstance(rec.record, Content1)
        self.assertEqual(rec.record.__id__, rowid)
        self.assertRaises(KeyError, model.__getitem__, 'add.html')
        self.assertRaises(KeyError, model.__getitem__, 'unknown')

    def test_model_remove_errors(self):
        from ptah.manage.model import ModelModule, ModelView

        mod = ModelModule(None, DummyRequest())
        model = mod['content1']

        form = ModelView(
            model, DummyRequest(
                POST=MultiDict([('form.buttons.remove', 'Remove')])))
        form.csrf = False
        form.update()

        self.assertIn("Please select records for removing.",
                      ptah.view.render_messages(form.request))

    def test_model_remove(self):
        from ptah.manage.model import ModelModule, ModelView

        content = Content1()
        content.title = 'Content test'

        Session = ptah.get_session()
        Session.add(content)
        Session.flush()

        rowid = content.__id__
        transaction.commit()

        mod = ModelModule(None, DummyRequest())
        model = mod['content1']

        form = ModelView(
            model, DummyRequest(
                POST=MultiDict(
                    list({'rowid':rowid,
                          'form.buttons.remove': 'Remove'}.items()))))
        form.csrf = False
        res = form.update()

        self.assertIsInstance(res, HTTPFound)

        transaction.commit()

        Session = ptah.get_session()
        rec = Session.query(Content1).filter(
            Content1.__id__ == rowid).first()
        self.assertIsNone(rec)

    def test_model_list(self):
        from ptah.manage.model import ModelModule, ModelView

        content = Content1()
        content.title = 'Content test'

        Session = ptah.get_session()
        Session.add(content)
        Session.flush()

        rowid = content.__id__
        transaction.commit()

        mod = ModelModule(None, DummyRequest())
        model = mod['content1']

        res = render_view_to_response(model, DummyRequest(), '', False)
        self.assertIn('value="%s"'%rowid, res.text)

        res = render_view_to_response(
            model, DummyRequest(params={'batch': 0}), '', False)
        self.assertIn('value="%s"'%rowid, res.text)

        res = render_view_to_response(
            model, DummyRequest(params={'batch': 'unknown'}), '', False)
        self.assertIn('value="%s"'%rowid, res.text)

    def test_model_add(self):
        from ptah.manage.model import ModelModule, ModelView

        mod = ModelModule(None, DummyRequest())
        model = mod['content1']

        form = ModelView(
            model, DummyRequest(POST={'form.buttons.add': 'Add'}))
        form.csrf = False
        res = form.update()

        self.assertIsInstance(res, HTTPFound)
        self.assertEqual(res.headers['location'], 'add.html')


class TestAddRecord(Base):

    def test_model_add_errors(self):
        from ptah.manage.model import ModelModule, AddRecord

        mod = ModelModule(None, DummyRequest())
        model = mod['content1']

        form = AddRecord(
            model, DummyRequest(POST={'form.buttons.add': 'Add'}))
        form.csrf = False
        form.update()

        self.assertIn("Please fix indicated errors.",
                      ptah.view.render_messages(form.request))

    def test_model_add_back(self):
        from ptah.manage.model import ModelModule, AddRecord

        mod = ModelModule(None, DummyRequest())
        model = mod['content1']

        form = AddRecord(
            model, DummyRequest(POST={'form.buttons.back': 'Back'}))
        form.csrf = False
        res = form.update()

        self.assertIsInstance(res, HTTPFound)
        self.assertEqual(res.headers['location'], '.')

    def test_model_add(self):
        from ptah.manage.model import ModelModule, AddRecord

        mod = ModelModule(None, DummyRequest())
        model = mod['content1']

        form = AddRecord(
            model, DummyRequest(
                POST={'title': 'Test content',
                      'form.buttons.add': 'Add'}))
        form.csrf = False
        res = form.update()

        id = form.record.__id__
        transaction.commit()

        Session = ptah.get_session()
        self.assertIsInstance(res, HTTPFound)
        self.assertEqual(res.headers['location'], './%s/'%id)
        self.assertEqual(len(Session.query(self.Content1).all()), 1)

        content = Session.query(self.Content1) \
            .filter(self.Content1.__id__ == id).first()
        self.assertEqual(content.title, 'Test content')

    def test_model_add_node(self):
        from ptah.manage.model import ModelModule, AddRecord

        mod = ModelModule(None, DummyRequest())
        model = mod['content2']

        form = AddRecord(
            model, DummyRequest(
                POST={'title': 'Test content', 'form.buttons.add': 'Add'}))
        form.csrf = False
        res = form.update()

        id = form.record.__id__
        transaction.commit()

        Session = ptah.get_session()
        self.assertIsInstance(res, HTTPFound)
        self.assertEqual(res.headers['location'], './%s/'%id)
        self.assertEqual(len(Session.query(self.Content2).all()), 1)

        content = Session.query(self.Content2) \
            .filter(self.Content2.__id__ == id).first()
        self.assertEqual(content.title, 'Test content')


class TestEditRecord(Base):

    def test_model_edit_back(self):
        from ptah.manage.model import ModelModule, EditRecord

        content = Content1()
        content.title = 'Content test'

        Session = ptah.get_session()
        Session.add(content)
        Session.flush()

        rowid = content.__id__
        transaction.commit()

        mod = ModelModule(None, DummyRequest())
        model = mod['content1'][rowid]

        form = EditRecord(
            model, DummyRequest(POST={'form.buttons.back': 'Back'}))
        form.csrf = False
        res = form.update()

        self.assertIsInstance(res, HTTPFound)
        self.assertEqual(res.headers['location'], '../')

    def test_model_edit_errors(self):
        from ptah.manage.model import ModelModule, EditRecord

        content = Content1()
        content.title = 'Content test'

        Session = ptah.get_session()
        Session.add(content)
        Session.flush()

        rowid = content.__id__
        transaction.commit()

        mod = ModelModule(None, DummyRequest())
        model = mod['content1'][rowid]

        form = EditRecord(
            model, DummyRequest(POST={'form.buttons.modify': 'Modify'}))
        form.csrf = False
        form.update()

        self.assertIn("Please fix indicated errors.",
                      ptah.view.render_messages(form.request))

    def test_model_edit(self):
        from ptah.manage.model import ModelModule, EditRecord

        content = Content1()
        content.title = 'Content test'

        Session = ptah.get_session()
        Session.add(content)
        Session.flush()

        rowid = content.__id__
        transaction.commit()

        mod = ModelModule(None, DummyRequest())
        model = mod['content1'][rowid]

        form = EditRecord(
            model, DummyRequest(
                POST={'title': 'Content', 'form.buttons.modify': 'Modify'}))
        form.csrf = False
        form.update()

        self.assertEqual(form.label, 'Record id: %s'%rowid)

        self.assertIn("Model record has been modified.",
                      ptah.view.render_messages(form.request))
        transaction.commit()

        Session = ptah.get_session()
        content = Session.query(Content1) \
            .filter(Content1.__id__ == rowid).first()
        self.assertEqual(content.title, 'Content')

    def test_model_edit_node(self):
        from ptah.manage.model import ModelModule, EditRecord

        content = Content2()
        content.title = 'Content test'

        Session = ptah.get_session()
        Session.add(content)
        Session.flush()

        rowid = content.__id__
        transaction.commit()

        mod = ModelModule(None, DummyRequest())
        model = mod['content2'][rowid]

        form = EditRecord(
            model, DummyRequest(
                POST={'title': 'Content', 'form.buttons.modify': 'Modify'}))
        form.csrf = False
        form.update()

        self.assertIn("Model record has been modified.",
                      ptah.view.render_messages(form.request))
        transaction.commit()

        Session = ptah.get_session()
        content = Session.query(Content2) \
            .filter(Content2.__id__ == rowid).first()
        self.assertEqual(content.title, 'Content')



""" type implementation """
import sys, logging
import sqlalchemy as sqla
from zope.interface import implementer
from pyramid.compat import string_types
from pyramid.threadlocal import get_current_registry

import ptah
from ptah import config
from ptah.cms.content import Content
from ptah.cms.container import BaseContainer
from ptah.cms.security import build_class_actions
from ptah.cms.interfaces import Forbidden, ContentSchema, ITypeInformation

log = logging.getLogger('ptah.cms')

TYPES_DIR_ID = 'ptah.cms:type'


@ptah.resolver('cms-type')
def typeInfoResolver(uri):
    """Type resolver

       :Parameters:
         - type scheme, e.g. blob-sql
       :Returns:
         - :py:class:`ptah.cms.TypeInformation`
    """
    return config.get_cfg_storage(TYPES_DIR_ID).get(uri)


def get_type(uri):
    """ Get registered type.

    :param uri: string identifier for TypeInformation, e.g. `cms-type:sqlblob`

    :Returns:
      - :py:class:`ptah.cms.TypeInformation`

    """
    return config.get_cfg_storage(TYPES_DIR_ID).get(uri)


def get_types():
    """ Get all registered types.

    :Returns:
      - mapping of all registered identifier and TypeInformation
    """
    return config.get_cfg_storage(TYPES_DIR_ID)


@implementer(ITypeInformation)
class TypeInformation(object):
    """ Type information """

    schema = None
    fieldset = None
    description = ''
    permission = ptah.NOT_ALLOWED

    addview = None # addview action, path relative to current container
    filter_content_types = False
    allowed_content_types = ()
    global_allow = True

    def __init__(self, cls, name, title, fieldset, **kw):
        self.__dict__.update(kw)

        self.__uri__ = 'cms-type:%s'%name

        self.cls = cls
        self.name = name
        self.title = title
        self.fieldset = fieldset

    def create(self, **data):
        content = self.cls(**data)
        get_current_registry().notify(ptah.events.ContentCreatedEvent(content))
        return content

    def is_allowed(self, container):
        if not isinstance(container, BaseContainer):
            return False

        if self.permission:
            return ptah.check_permission(self.permission, container)
        return True

    def check_context(self, container):
        if not self.is_allowed(container):
            raise Forbidden()

    def list_types(self, container):
        if container.__type__ is not self or \
               not isinstance(container, BaseContainer):
            return ()

        types = []
        all_types = config.get_cfg_storage(TYPES_DIR_ID)

        if self.filter_content_types:
            allowed_types = self.allowed_content_types
            if callable(allowed_types):
                allowed_types = allowed_types(container)

            for tinfo in allowed_types:
                if isinstance(tinfo, string_types):
                    tinfo = all_types.get('cms-type:%s'%tinfo)

                if tinfo and tinfo.is_allowed(container):
                    types.append(tinfo)
        else:
            for tinfo in all_types.values():
                if tinfo.global_allow and tinfo.is_allowed(container):
                    types.append(tinfo)

        return types


def Type(name, title=None, fieldset=None, **kw):
    """ Declare new type. This function has to be called within a content
    class declaration.

    .. code-block:: python

        class MyContent(ptah.cms.Content):

            __type__ = Type('My content')

    """
    info = config.DirectiveInfo(allowed_scope=('class',))

    fs = ContentSchema if fieldset is None else fieldset

    if title is None:
        title = name.capitalize()

    typeinfo = TypeInformation(None, name, title, fs, **kw)

    f_locals = sys._getframe(1).f_locals
    if '__mapper_args__' not in f_locals:
        f_locals['__mapper_args__'] = {'polymorphic_identity': typeinfo.__uri__}
    if '__id__' not in f_locals and '__tablename__' in f_locals:
        f_locals['__id__'] = sqla.Column(
            'id', sqla.Integer,
            sqla.ForeignKey('ptah_content.id'), primary_key=True)
    if '__uri_factory__' not in f_locals:
        schema = 'cms-{0}'.format(name)
        typeinfo.schema = schema

        f_locals['__uri_factory__'] = ptah.UriFactory(schema)

        def resolve_content(uri):
            return typeinfo.cls.__uri_sql_get__.first(uri=uri)

        resolve_content.__doc__ = 'CMS Content resolver for %s type'%title

        ptah.resolver(schema, 2)(resolve_content)

    # config actino and introspection info
    discr = (TYPES_DIR_ID, name)
    intr = config.Introspectable(TYPES_DIR_ID, discr, name, TYPES_DIR_ID)
    intr['name'] = name
    intr['type'] = typeinfo
    intr['codeinfo'] = info.codeinfo

    info.attach(
        config.ClassAction(
            register_type_impl, (typeinfo, name, fieldset), kw,
            discriminator=discr, introspectables=(intr,))
        )
    return typeinfo


excludeNames = ('expires', 'contributors', 'creators', 'view', 'subjects',
                'publisher', 'effective', 'created', 'modified')
def names_filter(name, fieldNames=None):
    if fieldNames is not None and name in fieldNames:
        return True

    if name in excludeNames:
        return False

    return not name.startswith('_')


def register_type_impl(
    config, cls, tinfo, name, fieldset,
    permission = ptah.NOT_ALLOWED, fieldNames=None, **kw):

    # generate schema
    if fieldset is None:
        fieldset = ptah.generate_fieldset(
            cls, fieldNames=fieldNames, namesFilter=names_filter)
        log.info("Generating fieldset for %s content type.", cls)

    if 'global_allow' not in kw and not issubclass(cls, Content):
        kw['global_allow'] = False

    tinfo.__dict__.update(kw)

    if fieldset is not None:
        tinfo.fieldset = fieldset

    tinfo.cls = cls
    tinfo.permission = permission

    config.get_cfg_storage(TYPES_DIR_ID)[tinfo.__uri__] = tinfo

    # sql query for content resolver
    cls.__uri_sql_get__ = ptah.QueryFreezer(
        lambda: ptah.get_session().query(cls) \
            .filter(cls.__uri__ == sqla.sql.bindparam('uri')))

    # build cms actions
    build_class_actions(cls)



""" password tool """
import translationstring
from os import urandom
from datetime import timedelta
from codecs import getencoder
from hashlib import sha1
from base64 import urlsafe_b64encode, urlsafe_b64decode
from pyramid.compat import bytes_

import ptah
from ptah import config, form, token

_ = translationstring.TranslationStringFactory('ptah')


ID_PASSWORD_CHANGER = 'ptah.password:changer'

TOKEN_TYPE = token.TokenType(
    '35c9b7df958f4e93ae9b275a7dc2219e', timedelta(minutes=10),
    'Reset password tokens')


class PlainPasswordManager(object):
    """PLAIN password manager."""

    def encode(self, password, salt=None):
        return '{plain}%s' % password

    def check(self, encoded, password):
        if encoded != password:
            return encoded == '{plain}%s' % password
        return True


class SSHAPasswordManager(object):
    """SSHA password manager."""

    _encoder = getencoder("utf-8")

    def encode(self, password, salt=None):
        if salt is None:
            salt = urandom(4)
        hash = sha1(self._encoder(password)[0])
        hash.update(salt)
        return bytes_('{ssha}','ascii') + urlsafe_b64encode(hash.digest()+salt)

    def check(self, encoded_password, password):
        # urlsafe_b64decode() cannot handle unicode input string. We
        # encode to ascii. This is safe as the encoded_password string
        # should not contain non-ascii characters anyway.
        encoded_password = bytes_(encoded_password, 'ascii')
        byte_string = urlsafe_b64decode(encoded_password[6:])
        salt = byte_string[20:]
        return encoded_password == self.encode(password, salt)


class PasswordTool(object):
    """ Password management utility. """

    pm = {'{plain}': PlainPasswordManager(),
          '{ssha}': SSHAPasswordManager(),
          }

    @property
    def manager(self):
        PWD_CONFIG = ptah.get_settings(ptah.CFG_ID_PTAH)
        try:
            return self.pm['{%s}' % PWD_CONFIG['pwd_manager']]
        except KeyError:
            return self.pm['{plain}']

    def check(self, encoded, password):
        """ Compare encoded password with plain password.

        :param encoded: Encoded password
        :param password: Plain password
        """
        try:
            pm, pwd = encoded.split('}', 1)
        except:
            return self.manager.check(encoded, password)

        manager = self.pm.get('%s}' % pm)
        if manager is not None:
            return manager.check(encoded, password)
        return False

    def encode(self, password, salt=None):
        """ Encode password with current password manager """
        return self.manager.encode(password, salt)

    def can_change_password(self, principal):
        """ Can principal password be changed.
        :py:class:`ptah.password_changer` is beeing used. """
        return ptah.extract_uri_schema(principal.__uri__) in \
            config.get_cfg_storage(ID_PASSWORD_CHANGER)

    def get_principal(self, passcode):
        """ Return principal by previously generated passcode. """
        data = token.service.get(passcode)

        if data is not None:
            return ptah.resolve(data)

    def generate_passcode(self, principal):
        """ Generate passcode for principal.

        :param principal: Principal object
        """
        return token.service.generate(TOKEN_TYPE, principal.__uri__)

    def remove_passcode(self, passcode):
        """ Remove passcode """
        token.service.remove(passcode)

    def change_password(self, passcode, password):
        """ Encode and change password.
        :py:class:`ptah.password_changer` is beeing used.

        :param passcode: Previously generated passcode
        :param passsword: Plain password.
        :rtype: True if password has been changed, False otherwise.
        """
        principal = self.get_principal(passcode)

        self.remove_passcode(passcode)

        if principal is not None:
            changers = config.get_cfg_storage(ID_PASSWORD_CHANGER)

            changer = changers.get(ptah.extract_uri_schema(principal.__uri__))
            if changer is not None:
                changer(principal, self.encode(password))
                return True

        return False

    def validate(self, password):
        """ Validate password """
        PWD_CONFIG = ptah.get_settings(ptah.CFG_ID_PTAH)

        if len(password) < PWD_CONFIG['pwd_min_length']:
            #return _('Password should be at least ${count} characters.',
            #         mapping={'count': self.min_length})
            return 'Password should be at least %s characters.' % \
                PWD_CONFIG['pwd_min_length']
        elif PWD_CONFIG['pwd_letters_digits'] and \
                (password.isalpha() or password.isdigit()):
            return _('Password should contain both letters and digits.')
        elif PWD_CONFIG['pwd_letters_mixed_case'] and \
                (password.isupper() or password.islower()):
            return _('Password should contain letters in mixed case.')


pwd_tool = PasswordTool()


class password_changer(object):
    """ Register password changer function.

    :param schema: Principal uri schema.

    .. code-block:: python

      @ptah.password_change('myuser')
      def change_password(principal, password):
          principal.password = password

    """
    def __init__(self, schema, __depth=1):
        self.info = config.DirectiveInfo(__depth)
        self.discr = (ID_PASSWORD_CHANGER, schema)

        self.intr = config.Introspectable(
            ID_PASSWORD_CHANGER, self.discr, schema, ID_PASSWORD_CHANGER)
        self.intr['schema'] = schema

    def __call__(self, changer, cfg=None):
        self.intr.title = changer.__doc__
        self.intr['callable'] = changer

        self.info.attach(
            config.Action(
                lambda config, schema, changer: \
                    config.get_cfg_storage(ID_PASSWORD_CHANGER).update(
                            {schema: changer}),
                (self.intr['schema'], changer),
                discriminator=self.discr, introspectables=(self.intr,)),
            cfg)
        return changer

    @classmethod
    def pyramid(cls, cfg, schema, changer):
        """ pyramid password changer registration directive.

        :param schema: Principal uri schema.
        :param changer: Function

        .. code-block:: python

          config = Configurator()
          config.include('ptah')

          config.ptah_password_changer('custom-schema', custom_changer)
        """
        cls(schema, 3)(changer, cfg)


def passwordValidator(field, value):
    """ password schema validator
    that uses password tool for additional checks"""
    if value is not ptah.form.null:
        err = pwd_tool.validate(value)
        if err is not None:
            raise form.Invalid(field, err)


def passwordSchemaValidator(field, appstruct):
    """ password schema validator that checks
    equality of password and confirm_password"""
    if appstruct['password'] and appstruct['confirm_password']:
        if appstruct['password'] != appstruct['confirm_password']:
            raise form.Invalid(
                field, _("Password and Confirm Password should be the same."))

        passwordValidator(field, appstruct['password'])


PasswordSchema = form.Fieldset(

    form.FieldFactory(
        'password',
        'password',
        title = _('Password'),
        description = _('Enter password. '\
                        'No spaces or special characters, should contain '\
                        'digits and letters in mixed case.'),
        default = ''),

    form.FieldFactory(
        'password',
        'confirm_password',
        title = _('Confirm password'),
        description = _('Re-enter the password. '
                        'Make sure the passwords are identical.'),
        default = ''),

    validator = passwordSchemaValidator
)



""" formatter tests """
import pytz
from datetime import datetime, timedelta
from pyramid.exceptions import ConfigurationConflictError

import ptah
from ptah.testing import PtahTestCase


class TestFormatterReg(PtahTestCase):

    _init_ptah = False

    def test_formatter_registration(self):
        @ptah.formatter('simple')
        def simple(v):
            return 'simple-%s'%v

        self.init_ptah()

        self.assertEqual(ptah.format.simple('test'), 'simple-test')
        self.assertIs(ptah.format.simple, simple)

    def test_formatter_registration_duplicate(self):
        @ptah.formatter('simple')
        def simple1(v):
            """ """

        @ptah.formatter('simple')
        def simple2(v):
            """ """

        self.assertRaises(ConfigurationConflictError, self.init_ptah)

    def test_formatter_introspector(self):
        @ptah.formatter('simple')
        def simple(v):
            """ doc """

        from ptah.formatter import ID_FORMATTER
        discr = (ID_FORMATTER, 'simple')

        self.init_ptah()

        intr = self.config.introspector.get(ID_FORMATTER, discr)

        self.assertIsNotNone(intr)
        self.assertEqual(intr['name'], 'simple')
        self.assertEqual(intr['description'], ' doc ')
        self.assertEqual(intr['callable'], simple)


class TestFormatter(PtahTestCase):

    def test_datetime_formatter(self):
        format = ptah.format

        # format only datetime
        self.assertEqual(format.datetime('text string'), 'text string')

        # default format
        dt = datetime(2011, 2, 6, 10, 35, 45, 80, pytz.UTC)
        self.assertEqual(format.datetime(dt, 'short'),
                         '02/06/11 04:35 AM')
        self.assertEqual(format.datetime(dt, 'medium'),
                         'Feb 06, 2011 04:35 AM')
        self.assertEqual(format.datetime(dt, 'long'),
                         'February 06, 2011 04:35 AM -0600')
        self.assertEqual(format.datetime(dt, 'full'),
                         'Sunday, February 06, 2011 04:35:45 AM CST')

    def test_datetime_formatter2(self):
        format = ptah.format

        # datetime without timezone
        dt = datetime(2011, 2, 6, 10, 35, 45, 80)
        self.assertEqual(format.datetime(dt, 'short'),
                         '02/06/11 04:35 AM')

        # different format
        dt = datetime(2011, 2, 6, 10, 35, 45, 80, pytz.UTC)

        FORMAT = ptah.get_settings('format', self.registry)
        FORMAT['date_short'] = '%b %d, %Y'

        self.assertEqual(format.datetime(dt, 'short'),
                         'Feb 06, 2011 04:35 AM')

    def test_timedelta_formatter(self):
        format = ptah.format

        # format only timedelta
        self.assertEqual(format.timedelta('text string'), 'text string')

        # full format
        td = timedelta(hours=10, minutes=5, seconds=45)

        self.assertEqual(format.timedelta(td, 'full'),
                         '10 hour(s) 5 min(s) 45 sec(s)')

        # medium format
        self.assertEqual(format.timedelta(td, 'medium'),
                         '10:05:45')

        # seconds format
        self.assertEqual(format.timedelta(td, 'seconds'),
                         '36345.0000')

        # default format
        self.assertEqual(format.timedelta(td), '10:05:45')

    def test_size_formatter(self):
        format = ptah.format

        # format only timedelta
        self.assertEqual(format.size('text string'), 'text string')

        v = 1024
        self.assertEqual(format.size(v, 'b'), '1024 B')

        self.assertEqual(format.size(v, 'k'), '1.00 KB')

        self.assertEqual(format.size(1024*768, 'm'), '0.75 MB')
        self.assertEqual(format.size(1024*768*768, 'm'), '576.00 MB')

        self.assertEqual(format.size(1024*768*768, 'g'), '0.56 GB')



import ptah
from ptah.testing import PtahTestCase
from pyramid.testing import DummyRequest
from pyramid.httpexceptions import HTTPFound
from pyramid.view import render_view_to_response


class TestSourceView(PtahTestCase):

    def test_source(self):
        from ptah.manage.source import SourceView
        from ptah.manage.manage import PtahManageRoute

        ptah.auth_service.set_userid(ptah.SUPERUSER_URI)

        request = DummyRequest()
        manage = PtahManageRoute(request)
        res = render_view_to_response(manage, request, 'source.html', False)
        self.assertIsInstance(res, HTTPFound)
        self.assertEqual(res.headers['location'], '.')

    def test_source_view(self):
        from ptah.manage.source import SourceView
        from ptah.manage.manage import PtahManageRoute

        ptah.auth_service.set_userid(ptah.SUPERUSER_URI)

        request = DummyRequest(
            params = {'pkg': 'ptah.config'})
        manage = PtahManageRoute(request)
        res = render_view_to_response(manage, request, 'source.html', False)
        self.assertIn('Source: ptah/config.py', res.text)

    def test_source_view_unknown(self):
        from ptah.manage.source import SourceView
        from ptah.manage.manage import PtahManageRoute

        ptah.auth_service.set_userid(ptah.SUPERUSER_URI)

        request = DummyRequest(
            params = {'pkg': 'unknown'})
        manage = PtahManageRoute(request)
        res = render_view_to_response(manage, request, 'source.html', False)
        self.assertIsInstance(res, HTTPFound)
        self.assertEqual(res.headers['location'], '.')



# ptah api

try:
    from collections import OrderedDict
except ImportError: # pragma: no cover
    import collections
    from ordereddict import OrderedDict
    collections.OrderedDict = OrderedDict

# config
from ptah import config
from ptah.config import adapter
from ptah.config import subscriber
from ptah.config import get_cfg_storage
from ptah.config import shutdown
from ptah.config import shutdown_handler

# uri
from ptah.uri import resolve
from ptah.uri import resolver
from ptah.uri import extract_uri_schema
from ptah.uri import UriFactory

# sqla
from ptah.sqlautils import sa_session
from ptah.sqlautils import get_base
from ptah.sqlautils import get_session
from ptah.sqlautils import get_session_maker
from ptah.sqlautils import reset_session

# events
from ptah import events
from ptah.events import event

# view api
from ptah.view import View
from ptah.view import add_message
from ptah.view import render_message
from ptah.view import render_messages

from ptah.view import snippet
from ptah.view import render_snippet

# layouts
from ptah.layout import layout
from ptah.layout import wrap_layout

# resource library
from ptah.library import library
from ptah.library import include
from ptah.library import render_includes

# settings
from ptah.settings import get_settings
from ptah.settings import register_settings
from ptah.settings import load_dbsettings

# security
from ptah.authentication import auth_service
from ptah.authentication import SUPERUSER_URI

from ptah.authentication import auth_checker
from ptah.authentication import auth_provider

from ptah.authentication import search_principals
from ptah.authentication import principal_searcher

# acl
from ptah.security import ACL
from ptah.security import ACLsProperty
from ptah.security import get_acls
from ptah.interfaces import IACLsAware

# role
from ptah.security import Role
from ptah.security import get_roles
from ptah.security import get_local_roles
from ptah.security import roles_provider
from ptah.interfaces import IOwnersAware
from ptah.interfaces import ILocalRolesAware

# permission
from ptah.security import Permission
from ptah.security import get_permissions
from ptah.security import check_permission

# default roles and permissions
from ptah.security import Everyone
from ptah.security import Authenticated
from ptah.security import Owner
from ptah.security import DEFAULT_ACL
from ptah.security import NOT_ALLOWED
from pyramid.security import ALL_PERMISSIONS
from pyramid.security import NO_PERMISSION_REQUIRED

# ptah settings ids
CFG_ID_PTAH = 'ptah'
CFG_ID_FORMAT = 'format'

# password tool
from ptah.password import pwd_tool
from ptah.password import password_changer

# formatter
from ptah.formatter import format, formatter

# rest api
from ptah.rest import RestService

# mail templates
from ptah import mail

# pagination
from ptah.util import Pagination

# thread local data
from ptah.util import tldata

# ReST renderer
from ptah.rst import rst_to_html

# sqlalchemy utils
from ptah.sqlautils import transaction
from ptah.sqlautils import QueryFreezer
from ptah.sqlautils import JsonDictType
from ptah.sqlautils import JsonListType
from ptah.sqlautils import set_jsontype_serializer
from ptah.sqla import generate_fieldset
from ptah.sqla import build_sqla_fieldset

# simple ui actions
from ptah.uiactions import uiaction
from ptah.uiactions import list_uiactions

# manage
from ptah import manage

# cms
from ptah import cms

# form api
from ptah import form

# populate
POPULATE = False
from ptah.populate import populate
from ptah.populate import POPULATE_DB_SCHEMA

# simple test case
from ptah.testing import PtahTestCase

# register migration
from ptah.migrate import register_migration

register_migration('ptah', 'ptah:migrations', 'Ptah database migration')


def includeme(cfg):
    # auth
    from ptah.security import PtahAuthorizationPolicy
    from pyramid.compat import configparser
    from pyramid.authentication import AuthTktAuthenticationPolicy

    kwargs = {'wild_domain': False,
              'callback': get_local_roles,
              'secret': cfg.registry.settings.get('ptah.secret','')}

    cfg.set_authorization_policy(PtahAuthorizationPolicy())
    cfg.set_authentication_policy(AuthTktAuthenticationPolicy(**kwargs))

    # include extra packages
    cfg.include('pyramid_tm')
    cfg.include('ptah.manage')

    # object events handler
    cfg.registry.registerHandler(
        config.ObjectEventNotify(cfg.registry), (config.IObjectEvent,))

    # initialize settings
    from ptah import settings
    from pyramid.interfaces import PHASE1_CONFIG
    def pyramid_init_settings(cfg, custom_settings=None,
                              section=configparser.DEFAULTSECT):
        cfg.action('ptah.init_settings',
                   settings.init_settings,
                   (cfg, custom_settings, section), order=999999)

    cfg.add_directive('ptah_init_settings', pyramid_init_settings)

    # initialize sql
    from ptah import ptahsettings
    cfg.add_directive('ptah_init_sql', ptahsettings.initialize_sql)

    # ptah manage ui directive
    cfg.add_directive('ptah_init_manage', ptahsettings.enable_manage)

    # ptah mailer directive
    cfg.add_directive('ptah_init_mailer', ptahsettings.set_mailer)

    # ptah rest api directive
    from ptah import rest
    cfg.add_directive('ptah_init_rest', rest.enable_rest_api)

    # ptah.config directives
    from ptah.config import pyramid_get_cfg_storage
    cfg.add_directive(
        'get_cfg_storage', pyramid_get_cfg_storage)

    # ptah.config.settings directives
    from ptah.settings import pyramid_get_settings
    cfg.add_directive(
        'ptah_get_settings', pyramid_get_settings)

    # ptah.authentication directives
    from ptah import authentication
    cfg.add_directive(
        'ptah_auth_checker', authentication.pyramid_auth_checker)
    cfg.add_directive(
        'ptah_auth_provider', authentication.auth_provider.pyramid)
    cfg.add_directive(
        'ptah_principal_searcher', authentication.principal_searcher.pyramid)

    # ptah.uri directives
    cfg.add_directive('ptah_uri_resolver', resolver.pyramid)

    # ptah.password directives
    from ptah import password
    cfg.add_directive('ptah_password_changer', password_changer.pyramid)

    # layout directive
    cfg.add_directive('ptah_layout', layout.pyramid)

    # snippet directive
    cfg.add_directive('ptah_snippet', snippet.pyramid)

    # populate
    def pyramid_populate(cfg):
        from ptah.populate import Populate
        if not POPULATE:
            cfg.action('ptah.populate',
                       Populate(cfg.registry).execute, order=9999999)

    cfg.add_directive('ptah_populate', pyramid_populate)
    cfg.add_directive('ptah_populate_step', populate.pyramid)

    # migrations
    from ptah import migrate
    cfg.add_directive('ptah_migrate', migrate.ptah_migrate)

    # ptah static assets
    cfg.add_static_view('_ptah/static', 'ptah:static/')

    # scan ptah
    cfg.scan('ptah')

    cfg.add_translation_dirs('ptah:locale')



import transaction
import sqlalchemy as sqla
import ptah
from ptah.testing import PtahTestCase


class TestSqlSession(PtahTestCase):

    def test_transaction(self):
        from ptah.sqlautils import transaction

        class SA(object):

            commited = False
            rollbacked = False

            raise_commit = False

            def commit(self):
                if self.raise_commit:
                    raise RuntimeError()
                self.commited = True

            def rollback(self):
                self.rollbacked = True

        sa = SA()
        trans = transaction(sa)

        with trans:
            pass

        self.assertTrue(sa.commited)

        sa = SA()
        trans = transaction(sa)
        try:
            with trans:
                raise ValueError()
        except:
            pass
        self.assertTrue(sa.rollbacked)

        sa = SA()
        sa.raise_commit = True
        trans = transaction(sa)
        try:
            with trans:
                pass
        except:
            pass
        self.assertTrue(sa.rollbacked)

    def test_sa_session(self):
        with ptah.sa_session() as sa:
            self.assertIs(sa, ptah.get_session())

    def test_sa_session_nested(self):
        err = None

        try:
            with ptah.sa_session() as sa:
                with ptah.sa_session() as sa:
                    pass
        except Exception as e:
            err = e

        self.assertIsNotNone(err)



import sys
import ptah
from ptah.scripts import manage
from pyramid.compat import NativeIO


class TestManageCommand(ptah.PtahTestCase):

    _init_ptah = False

    def test_manage_no_params(self):
        self.init_ptah()

        sys.argv[:] = ['ptah-manage', 'ptah.ini']

        stdout = sys.stdout
        out = NativeIO()
        sys.stdout = out

        manage.main(False)
        sys.stdout = stdout

        val = out.getvalue()

        self.assertIn(
            '[-h] [--list-modules] [--list-models] config', val)

    def test_list_modules(self):

        @ptah.manage.module('custom')
        class CustomModule(ptah.manage.PtahModule):
            """ Custom module description """

            title = 'Custom Module'

        self.init_ptah()

        sys.argv[1:] = ['--list-modules', 'ptah.ini']

        stdout = sys.stdout
        out = NativeIO()
        sys.stdout = out

        manage.main(False)
        sys.stdout = stdout

        val = out.getvalue()

        self.assertIn('* custom: Custom Module (disabled: False)', val)
        self.assertIn('Custom module description', val)

        # disable
        cfg = ptah.get_settings(ptah.CFG_ID_PTAH)
        cfg['disable_modules'] = ('custom',)

        out = NativeIO()
        sys.stdout = out

        manage.main(False)
        sys.stdout = stdout

        val = out.getvalue()

        self.assertIn('* custom: Custom Module (disabled: True)', val)

    def test_list_models(self):
        global CustomModel

        class CustomModel(ptah.cms.Content):
            """ Custom module description """

            title = 'Custom Module'

            __type__ = ptah.cms.Type(
                'custom', title='Custom model',
                description = 'Custom model description')

        self.init_ptah()

        sys.argv[1:] = ['--list-models', 'ptah.ini']

        stdout = sys.stdout
        out = NativeIO()
        sys.stdout = out

        manage.main(False)
        sys.stdout = stdout

        val = out.getvalue()

        self.assertIn('* cms-type:custom: Custom model (disabled: False)', val)
        self.assertIn('Custom model description', val)
        self.assertIn('class: CustomModel', val)
        self.assertIn('module: test_manage', val)

        # disable
        cfg = ptah.get_settings(ptah.CFG_ID_PTAH)
        cfg['disable_models'] = ('cms-type:custom',)

        out = NativeIO()
        sys.stdout = out

        manage.main(False)
        sys.stdout = stdout

        val = out.getvalue()

        self.assertIn('* cms-type:custom: Custom model (disabled: True)', val)



""" settings module """
import ptah
from ptah.settings import ID_SETTINGS_GROUP
from pyramid.decorator import reify
from pyramid.view import view_config
from pyramid.httpexceptions import HTTPFound


@ptah.manage.module('settings')
class SettingsModule(ptah.manage.PtahModule):
    __doc__ = 'The current settings which include defaults not used by .ini'

    title = 'Settings'

    def __getitem__(self, key):
        grp = ptah.get_settings(key, self.request.registry)
        if grp is not None and grp.__ttw__:
            return SettingsWrapper(grp, self)
        raise KeyError(key)


class SettingsWrapper(object):

    def __init__(self, grp, mod):
        self.__name__ = grp.__name__
        self.__parent__ = mod
        self.group = grp


@view_config(
    context=SettingsModule, wrapper = ptah.wrap_layout(),
    renderer='ptah.manage:templates/settings.pt')

class SettingsView(ptah.View):
    """ Settings manage module view """

    grps = None

    def update(self):
        groups = ptah.get_cfg_storage(ID_SETTINGS_GROUP).items()

        data = []
        for name, group in sorted(groups):
            if self.grps is not None and name not in self.grps:
                continue

            title = group.__title__ or name
            description = group.__description__

            schema = []
            for field in group.__fields__.values():
                if getattr(field, 'tint', False):
                    value = '* * * * * * *'
                else:
                    value = field.dumps(group[field.name])
                schema.append(
                    ({'name': '{0}.{1}'.format(name, field.name),
                      'type': field.__class__.__name__,
                      'value': value,
                      'title': field.title,
                      'description': field.description,
                      'default': field.dumps(field.default)}))

            data.append(
                ({'title': title,
                  'description': description,
                  'schema': schema,
                  'name': group.__name__,
                  'ttw': group.__ttw__}))

        return {'data': sorted(data, key=lambda item: item['title'])}


@view_config(context=SettingsWrapper, wrapper=ptah.wrap_layout())
class EditForm(ptah.form.Form):
    """ Settings group edit form """

    @reify
    def label(self):
        return self.context.group.__title__

    @reify
    def description(self):
        return self.context.group.__description__

    @reify
    def fields(self):
        grp = self.context.group
        return grp.__fields__.omit(*grp.__ttw_skip_fields__)

    def form_content(self):
        return self.context.group

    @ptah.form.button('Modify', actype=ptah.form.AC_PRIMARY)
    def modify_handler(self):
        data, errors = self.extract()
        if errors: # pragma: no cover
            self.message(errors, 'form-error')
            return

        self.message("Settings have been modified.")
        self.context.group.updatedb(**data)
        return HTTPFound('../#{0}'.format(self.context.group.__name__))

    #@ptah.form.button('Reset defaults', actype=ptah.form.AC_INFO)
    #def reset_handler(self):
    #    pass

    @ptah.form.button('Back')
    def back_handler(self):
        return HTTPFound('..')



""" layout implementation """
import json
import logging
import random
import inspect
from collections import namedtuple
from collections import OrderedDict
from zope.interface import providedBy, Interface

from pyramid.compat import text_, string_types
from pyramid.config.views import DefaultViewMapper
from pyramid.httpexceptions import HTTPException
from pyramid.location import lineage
from pyramid.renderers import RendererHelper
from pyramid.interfaces import IRequest
from pyramid.interfaces import IRouteRequest
from pyramid.interfaces import IView
from pyramid.interfaces import IViewClassifier

from ptah import config
from ptah.view import View

log = logging.getLogger('ptah')

LAYOUT_ID = 'ptah.view:layout'
LAYOUT_WRAPPER_ID = 'ptah.view:layout-wrapper'

LayoutInfo = namedtuple(
    'LayoutInfo', 'name layout view original renderer intr')


class ILayout(Interface):
    """ marker interface """


def query_layout(root, context, request, name=''):
    """ query named layout for context """
    assert IRequest.providedBy(request), "must pass in a request object"

    try:
        iface = request.request_iface
    except AttributeError:
        iface = IRequest

    root = providedBy(root)

    adapters = request.registry.adapters

    for context in lineage(context):
        layout_factory = adapters.lookup(
            (root, iface, providedBy(context)), ILayout, name=name)

        if layout_factory is not None:
            return layout_factory, context

    return None, None


def query_layout_chain(root, context, request, layoutname=''):
    chain = []

    layout, layoutcontext = query_layout(root, context, request, layoutname)
    if layout is None:
        return chain

    chain.append((layout, layoutcontext))
    contexts = {layoutname: layoutcontext}

    while layout is not None and layout.layout is not None:
        if layout.layout in contexts:
            l_context = contexts[layout.layout].__parent__
        else:
            l_context = context

        layout, layoutcontext = query_layout(
            root, l_context, request, layout.layout)
        if layout is not None:
            chain.append((layout, layoutcontext))
            contexts[layout.name] = layoutcontext

            if layout.layout is None:
                break

    return chain


class layout(object):
    """Registers a layout.

    :param name: Layout name
    :param context: Specific context for this layout.
    :param root:  Root object
    :param parent: A parent layout. None means no parent layout.
    :param renderer: A pyramid renderer
    :param route_name: A pyramid route_name. Apply layout only for
        specific route
    :param use_global_views: Apply layout to all routes. even is route
        doesnt use use_global_views.


    Simple example with one default layout and 'page' layout.

    .. code-block:: python

      import ptah

      @ptah.layout('page', parent='page', renderer='ptah:template/page.pt')
      class PageLayout(ptah.View):
           ...

      @ptah.layout('', parent='page', renderer='ptah:template/template.pt')
      class DefaultLayout(object):
           ...

    To use layout with pyramid view use ``wrapper=ptah.wrap_layout()``

    Example:

    .. code-block:: python

      config.add_view('
          index.html',
          wrapper=ptah.wrap_layout(),
          renderer = '...')

    in this example '' layout is beeing used. You can specify specific layout
    name for pyramid view ``ptah.wrap_layout('page')``

    """

    def __init__(self, name='', context=None, root=None, parent=None,
                 renderer=None, route_name=None, use_global_views=False,
                 __depth=1):
        self.info = config.DirectiveInfo(__depth)
        self.discr = (LAYOUT_ID, name, context, route_name)

        self.intr = intr = config.Introspectable(
            LAYOUT_ID, self.discr, name, LAYOUT_ID)

        intr['name'] = name
        intr['context'] = context
        intr['root'] = root
        intr['renderer'] = renderer
        intr['route_name'] = route_name
        intr['parent'] = parent
        intr['use_global_views'] = use_global_views
        intr['codeinfo'] = self.info.codeinfo

    @classmethod
    def register(cls, name='', context=None, root=None, parent='',
                 renderer=None, route_name=None, use_global_views=False,
                 view=View):
        """ Imperative layout registration.

        :param name: Layout name
        :param context: Specific context for this layout.
        :param root:  Root object
        :param parent: A parent layout. None means no parent layout.
        :param renderer: A pyramid renderer
        :param route_name: A pyramid route_name. Apply layout only for
            specific route
        :param use_global_views: Apply layout to all routes. even is route
            doesnt use use_global_views.
        :param view: Layout implementation (same as for pyramid view)

        .. code-block:: python

            ptah.layout.register('page', renderer='...', view=MyLayout)

        """
        layout(name, context, root, parent,
               renderer, route_name, use_global_views, 2)(view)

    @classmethod
    def pyramid(cls, cfg, name='', context=None, root=None, parent='',
                renderer=None, route_name=None,
                use_global_views=False, view=View):
        """ Pyramid `ptah_layout` directive:

        .. code-block:: python

          config = Configurator()
          config.include('ptah')

          config.ptah_layout('page', renderer='..')
        """
        l = layout(name, context, root, parent,
               renderer, route_name, use_global_views, 3)(view, cfg)

    def __call__(self, view, cfg=None):
        intr = self.intr
        intr['view'] = view

        self.info.attach(
            config.Action(
                self._register,
                discriminator=self.discr, introspectables=(intr,)),
            cfg)
        return view

    def _register(self, cfg):
        intr = self.intr

        view, name, context, root, \
              renderer, layout, route_name, use_global_views = \
              (intr['view'], intr['name'], intr['context'], intr['root'],
               intr['renderer'], intr['parent'], intr['route_name'],
               intr['use_global_views'])

        if not layout:
            layout = None
        elif layout == '.':
            layout = ''

        if isinstance(renderer, string_types):
            renderer = RendererHelper(name=renderer, registry=cfg.registry)

        request_iface = IRequest
        if route_name is not None:
            request_iface = cfg.registry.getUtility(
                IRouteRequest, name=route_name)

        if use_global_views:
            request_iface = Interface

        if context is None:
            context = Interface

        mapper = getattr(view, '__view_mapper__', DefaultViewMapper)
        mapped_view = mapper()(view)

        info = LayoutInfo(
            name, layout, mapped_view, view, renderer, intr)
        cfg.registry.registerAdapter(
            info, (root, request_iface, context), ILayout, name)


class LayoutRenderer(object):

    def __init__(self, layout):
        self.layout = layout

    def layout_info(self, layout, context, request, content,
                    colors=('green','blue','yellow','gray','black')):
        intr = layout.intr
        view = intr['view']

        data = OrderedDict(
            (('name', intr['name']),
             ('parent-layout', intr['parent']),
             ('layout-factory', '%s.%s'%(view.__module__, view.__name__)),
             ('python-module', intr['codeinfo'].module),
             ('python-module-location', intr['codeinfo'].filename),
             ('python-module-line', intr['codeinfo'].lineno),
             ('renderer', intr['renderer']),
             ('context', '%s.%s'%(context.__class__.__module__,
                                  context.__class__.__name__)),
             ('context-path', request.resource_url(context)),
             ))

        content = text_('\n<!-- layout:\n%s \n-->\n'\
                        '<div style="border: 2px solid %s">%s</div>')%(
            json.dumps(data, indent=2), random.choice(colors), content)

        return content

    def view_info(self, discr, context, request, content):
        introspector = request.registry.introspector

        template = 'unknown'
        intr = introspector.get('templates', discr)
        if intr is not None: # pragma: no cover
            template = intr['name']

        intr = introspector.get('views', discr)
        if intr is None: # pragma: no cover
            return content

        view = intr['callable']

        data = OrderedDict(
            (('name', intr['name']),
             ('route-name', intr['route_name']),
             ('view-factory', '%s.%s'%(view.__module__, view.__name__)),
             ('python-module', inspect.getmodule(view).__name__),
             ('python-module-location', inspect.getsourcefile(view)),
             ('python-module-line', inspect.getsourcelines(view)[-1]),
             ('renderer', template),
             ('context', '%s.%s'%(context.__class__.__module__,
                                  context.__class__.__name__)),
             ('context-path', request.resource_url(context)),
             ))

        content = text_('\n<!-- view:\n%s \n-->\n'\
                        '<div style="border: 2px solid red">%s</div>')%(
            json.dumps(data, indent=2), content)

        return content

    def __call__(self, context, request):
        chain = query_layout_chain(request.root, context, request, self.layout)
        if not chain:
            log.warning("Can't find layout '%s' for context '%s'",
                        self.layout, context)

        if isinstance(request.wrapped_response, HTTPException):
            return request.wrapped_response

        debug = getattr(request, '__layout_debug__', False)

        content = text_(request.wrapped_body, 'utf-8')

        if debug:
            content = self.view_info(debug, context, request, content)

        for layout, layoutcontext in chain:
            value = layout.view(layoutcontext, request)
            if value is None:
                value = {}

            system = {'view': request.__view__,
                      'renderer_info': layout.renderer,
                      'context': layoutcontext,
                      'request': request,
                      'wrapped_content': content}

            content = layout.renderer.render(value, system, request)

            if debug:
                content = self.layout_info(
                    layout, layoutcontext, request, content)

        request.response.text = content
        return request.response


def wrap_layout(layout=''):
    """ Generate view name for pyramid view declaration.

    .. code-block:: python

      config = Configurator()
      config.include('ptah')

      config.ptah_layout('page')

      config.add_view(
          'index.html',
          wrapper=ptah.wrap_layout())

    """
    name = '#layout-{0}'.format(layout)

    info = config.DirectiveInfo()
    info.attach(config.Action(wrap_layout_impl, (name, layout)))
    return name


def wrap_layout_impl(cfg, name, layout):
    renderer = cfg.registry.adapters.lookup(
        (IViewClassifier, Interface, Interface), IView, name=name)
    if renderer is None:
        cfg.registry.registerAdapter(
            LayoutRenderer(layout),
            (IViewClassifier, Interface, Interface), IView, name)



import pytz
from ptah.testing import PtahTestCase


class TestTimezoneField(PtahTestCase):

    def test_timezone_loads(self):
        from ptah.form import null, TimezoneField

        typ = TimezoneField('test')

        tz = typ.loads('"asia/almaty"')
        self.assertEqual(tz, pytz.timezone('Asia/Almaty'))

        tz = typ.loads('asia/almaty')
        self.assertEqual(tz, pytz.timezone('Asia/Almaty'))

        tz = typ.loads('Asia/Almaty')
        self.assertEqual(tz, pytz.timezone('Asia/Almaty'))


    def test_timezone_schema_serialize(self):
        from ptah.form import null, TimezoneField

        typ = TimezoneField('test')

        self.assertTrue(typ.serialize(null) is null)
        self.assertEqual(typ.serialize(pytz.UTC), 'utc')

    def test_timezone_schema_deserialize(self):
        from ptah.form import null, Invalid, TimezoneField

        typ = TimezoneField('test')

        self.assertTrue(typ.deserialize(null) is null)
        self.assertTrue(typ.deserialize('') is null)

        # special case for 'GMT+X' timezones
        self.assertEqual(repr(typ.deserialize('GMT+6')),
                         "<StaticTzInfo 'Etc/GMT+6'>")
        self.assertEqual(repr(typ.deserialize('gmt+6')),
                         "<StaticTzInfo 'Etc/GMT+6'>")

        # general timezones
        self.assertEqual(repr(typ.deserialize('US/Central')),
                         "<DstTzInfo 'US/Central' CST-1 day, 18:00:00 STD>")

        self.assertEqual(repr(typ.deserialize('us/central')),
                         "<DstTzInfo 'US/Central' CST-1 day, 18:00:00 STD>")

        # unknown timezone
        self.assertRaises(Invalid, typ.deserialize, 'unknown')



"""Ptah 0.3.x changes

Revision ID: 0301
Revises: None
Create Date: 2012-01-10 10:53:51.715491

"""

# revision identifiers, used by Alembic.
revision = '0301'
down_revision = None

import ptah
from alembic import op
from alembic import context
import sqlalchemy as sa
from sqlalchemy.engine import reflection


def upgrade():
    # it possible that this script is being called on new database
    # in this case we should not do anything
    insp = reflection.Inspector.from_engine(ptah.get_base().metadata.bind)

    if 'annotations' in [r['name'] for r in insp.get_columns('ptah_nodes')]:
        return

    # ptah_nodes
    op.add_column(
        'ptah_nodes',
        sa.Column('annotations', ptah.JsonDictType(), default={}))

    # ptah_content
    op.add_column(
        'ptah_content', sa.Column('lang', sa.String(12), default='en'))

    op.create_index(
        'ix_ptah_content_path', 'ptah_content', ('path',))

    # sqlite doesnt support column drop
    impl = context.get_impl()

    if impl.__dialect__ != 'sqlite':
        op.drop_column('ptah_content', 'view')
        op.drop_column('ptah_content', 'creators')
        op.drop_column('ptah_content', 'subjects')
        op.drop_column('ptah_content', 'publisher')
        op.drop_column('ptah_content', 'contributors')

    op.drop_table('test_sqla_table')


def downgrade():
    pass



import transaction
from datetime import timedelta
from pyramid.exceptions import ConfigurationConflictError

from ptah.testing import PtahTestCase


class TestTokenType(PtahTestCase):

    _init_ptah = False

    def test_token(self):
        from ptah import token

        tt = token.TokenType('unique-id', timedelta(minutes=20))
        self.init_ptah()

        t = token.service.generate(tt, 'data')
        transaction.commit()

        self.assertEqual(token.service.get(t), 'data')
        self.assertEqual(token.service.get_bydata(tt, 'data'), t)

        token.service.remove(t)
        self.assertEqual(token.service.get(t), None)

    def test_token_type(self):
        from ptah import token

        token.TokenType('unique-id', timedelta(minutes=20))
        token.TokenType('unique-id', timedelta(minutes=20))

        self.assertRaises(ConfigurationConflictError, self.init_ptah)

    def test_token_remove_expired(self):
        pass



""" introspector renderers """
import sys
import inspect
from pyramid.renderers import render
from pyramid.interfaces import IIntrospectable

import ptah
from ptah import config, manage, form
from ptah.manage import get_manage_url
from ptah.form.field import PREVIEW_ID


class Renderer(ptah.View):

    title = ''
    rst_to_html = staticmethod(ptah.rst_to_html)

    def __call__(self):
        return {'manage_url': get_manage_url(self.request)}


@ptah.snippet(
    ptah.event.ID_EVENT, IIntrospectable,
    renderer='ptah.manage:templates/intr-event.pt')
class EventRenderer(Renderer):
    """ List of event declarations """

    title = 'Events'


@ptah.snippet(
    'ptah.config:subscriber', IIntrospectable,
    renderer='ptah.manage:templates/intr-subscriber.pt')
class SubscriberRenderer(Renderer):
    """ List of event subscribers """

    title = 'Event subscribers'

    def getInfo(self):
        intr = self.context
        handler = intr['handler']
        required = intr['required']
        factoryInfo = '%s.%s'%(intr['codeinfo'].module, handler.__name__)

        if len(required) > 1: # pragma: no cover
            obj = required[0]
            klass = required[1]
            #event = config.get_cfg_storage(config.ID_EVENT).get(
            #    action.args[1][-1], None)
        else:
            obj = None
            klass = required[0]
            #event = config.get_cfg_storage(config.ID_EVENT).get(
            #    action.args[1][0], None)

        return locals()


@ptah.snippet(
    'ptah:permission', IIntrospectable,
    renderer='ptah.manage:templates/intr-permission.pt')
class PermissionRenderer(Renderer):
    """ Permission registrations """

    title = 'Permission'


@ptah.snippet(
    'ptah:role', IIntrospectable,
    renderer='ptah.manage:templates/intr-role.pt')
class RoleIntrospection(Renderer):
    """ Role registrations """

    title = 'Role'


@ptah.snippet(
    'ptah:resolver', IIntrospectable,
    renderer='ptah.manage:templates/intr-uriresolver.pt')
class UriRenderer(Renderer):
    """ Uri resolvers """

    title = 'Uri resolver'


@ptah.snippet(
    'ptah.manage:module', IIntrospectable,
    renderer='ptah.manage:templates/intr-managemodule.pt')
class ManageModuleRenderer(Renderer):
    """ List of registered ptah manage modules """

    title = 'Ptah manage modules'


@ptah.snippet(
    'ptah:aclmap', IIntrospectable,
    renderer='ptah.manage:templates/intr-aclmap.pt')
class ACLMapRenderer(Renderer):
    """ ACL map registration """

    title = 'ACL Maps'


@ptah.snippet(
    'ptah:authchecker', IIntrospectable,
    renderer='ptah.manage:templates/intr-authchecker.pt')
class AuthCheckerRenderer(Renderer):
    """ List of registered authentication checkers """

    title = 'Authentication checkers'


@ptah.snippet(
    'ptah:authprovider', IIntrospectable,
    renderer='ptah.manage:templates/intr-authprovider.pt')
class AuthProviderRenderer(Renderer):
    """ List of registered authentication providers """

    title = 'Authentication providers'


@ptah.snippet(
    'ptah:settings-group', IIntrospectable,
    renderer='ptah.manage:templates/intr-settings.pt')
class SettingsGroupRenderer(Renderer):
    """ List of registered settings groups """

    title = 'Settings'


@ptah.snippet(
    'ptah.form:field', IIntrospectable,
    renderer='ptah.manage:templates/intr-field.pt')
class FieldRenderer(Renderer):
    """ List of registered fields """

    title = 'Fields'

    def __call__(self):
        return {'manage_url': get_manage_url(self.request),
                'previews': config.get_cfg_storage(PREVIEW_ID)}


@ptah.snippet(
    'ptah:uiaction', IIntrospectable,
    renderer='ptah.manage:templates/intr-uiaction.pt')
class uiActionRenderer(Renderer):
    """ List of registered ui actions """

    title = 'UI Actions'


@ptah.snippet(
    'ptah:token-type', IIntrospectable,
    renderer='ptah.manage:templates/intr-tokentype.pt')
class TokenTypeRenderer(Renderer):
    """ List of registered token types """

    title = 'Token types'


@ptah.snippet(
    'ptah.cms:type', IIntrospectable,
    renderer='ptah.manage:templates/intr-contenttype.pt')
class ContentTypeRenderer(Renderer):
    """ Ptah content types """

    title = 'Content Types'


@ptah.snippet(
    'ptah:formatter', IIntrospectable,
    renderer='ptah.manage:templates/intr-formatter.pt')
class formatterRenderer(Renderer):
    """ List of formatters """

    title = 'Formatters'



import sys
import ptah
from ptah.scripts import settings
from ptah.testing import PtahTestCase
from pyramid.compat import NativeIO


class TestCommand(PtahTestCase):

    _init_ptah = False

    def test_settings_command(self):
        field = ptah.form.TextField(
            'node',
            default = 'test')

        ptah.register_settings(
            'group1', field,
            title = 'Section1',
            description = 'Description1',
            )

        ptah.register_settings(
            'group2', field,
            title = 'Section2',
            description = 'Description2',
            )

        self.init_ptah()

        group1 = ptah.get_settings('group1', self.registry)
        group2 = ptah.get_settings('group2', self.registry)

        # all
        sys.argv[1:] = ['-a', 'ptah.ini']

        stdout = sys.stdout
        out = NativeIO()
        sys.stdout = out

        settings.main(False)
        sys.stdout = stdout

        val = out.getvalue()
        self.assertIn('Section1', val)
        self.assertIn('Section2', val)
        self.assertIn('group1.node', val)
        self.assertIn('group2.node', val)

        # section
        sys.argv[1:] = ['-l', 'group1', 'ptah.ini']

        stdout = sys.stdout
        out = NativeIO()
        sys.stdout = out

        settings.main(False)
        sys.stdout = stdout

        val = out.getvalue()
        self.assertIn('Section1', val)
        self.assertNotIn('Section2', val)
        self.assertIn('group1.node', val)
        self.assertNotIn('group2.node', val)

        # print
        sys.argv[1:] = ['-p', 'ptah.ini']

        stdout = sys.stdout
        out = NativeIO()
        sys.stdout = out

        settings.main(False)
        sys.stdout = stdout

        val = out.getvalue().strip()
        self.assertIn('group1.node = "test"', val)
        self.assertIn('group2.node = "test"', val)



import inspect
from zope.interface import providedBy
from pyramid.view import view_config
from pyramid.interfaces import IView, IViewClassifier
from pyramid.httpexceptions import HTTPNotFound


@view_config('layout-preview.html')
def layoutPreview(context, request):
    view_name = request.GET.get('view', '')

    adapters = request.registry.adapters

    view = adapters.lookup(
        (IViewClassifier, providedBy(request), providedBy(context)),
        IView, name=view_name, default=None)

    if view is None:
        return HTTPNotFound()

    request.__layout_debug__ = view.__discriminator__(context, request)

    return view(context, request)



""" introspect module """
import urllib
from zope.interface import Interface
from pyramid.view import view_config
from pyramid.compat import string_types, url_unquote
from pyramid.interfaces import IIntrospectable

import ptah
from ptah.view import ISnippet
from ptah.manage import get_manage_url


@ptah.manage.module('introspect')
class IntrospectModule(ptah.manage.PtahModule):
    __doc__ = 'Insight into all configuration and registrations.'

    title = 'Introspect'

    def __getitem__(self, key):
        key = url_unquote(key)
        if key in self.request.registry.introspector.categories():
            return Introspector(key, self, self.request)

        raise KeyError(key)


class Introspector(object):

    def __init__(self, name, mod, request):
        self.__name__ = name
        self.__parent__ = mod

        self.name = name
        self.request = request


@view_config(
    context=IntrospectModule, wrapper=ptah.wrap_layout(),
    renderer='ptah.manage:templates/introspect.pt')
class MainView(ptah.View):
    __doc__ = 'Introspection module view.'

    def update(self):
        self.manage_url = '{0}/introspect'.format(get_manage_url(self.request))

        self.renderers = sorted(
            (self.request.registry.adapters.lookupAll(
                (IIntrospectable, Interface), ISnippet)),
            key = lambda item: item[1].view.title)


@view_config(
    context=Introspector, wrapper=ptah.wrap_layout(),
    renderer='ptah.manage:templates/introspect-intr.pt')
class IntrospectorView(ptah.View):

    def update(self):
        name = self.context.name
        registry = self.request.registry

        self.renderer = registry.adapters.lookup(
            (IIntrospectable, Interface), ISnippet, name=name)

        self.intrs = sorted(
            (item['introspectable'] for item in
             registry.introspector.get_category(name)),
            key = lambda item: item.title)

        self.manage_url = '{0}/introspect'.format(get_manage_url(self.request))

        self.renderers = sorted(
            (registry.adapters.lookupAll(
                (IIntrospectable, Interface), ISnippet)),
            key = lambda item: item[1].view.title)



""" assets libraries """
import ptah

# jQuery library
ptah.library(
    'jquery',
    path='ptah:static/jquery/jquery-1.7.1.min.js',
    type="js")

ptah.library(
    'jquery-ui',
    path='ptah:static/jquery/jquery-ui-1.8.18.min.js',
    type="js",
    require="jquery")

ptah.library(
    'jquery-ui',
    'ptah:static/jquery/jquery-ui-1.8.18.css',
    type='css')

# Bootstrap css
ptah.library(
    'bootstrap',
    path='ptah:static/bootstrap/bootstrap-2.0.2.min.css',
    type="css")

ptah.library(
    'bootstrap-js',
    path='ptah:static/bootstrap/bootstrap-2.0.2.min.js',
    type="js",
    require="jquery")

# CKEditor
ptah.library(
    'ckeditor',
    path=('ptah:static/ckeditor/ckeditor.js',
          'ptah:static/ckeditor/adapters/jquery.js'),
    type="js",
    require="jquery")



""" sqla module """
import urllib
from sqlalchemy.orm.mapper import _mapper_registry
from pyramid.view import view_config
from pyramid.compat import url_quote_plus
from pyramid.decorator import reify
from pyramid.httpexceptions import HTTPFound

import ptah
from ptah import form

Session = ptah.get_session()


@ptah.manage.module('sqla')
class SQLAModule(ptah.manage.PtahModule):
    __doc__ = 'A listing of all tables with ability to view and edit records'

    title = 'SQLAlchemy'

    metadata = {}

    def __getitem__(self, key):
        try:
            id, table = key.split('-', 1)
        except:
            raise KeyError(key)

        md = self.metadata[id][0]
        return Table(md.tables[table], self, self.request)

    def available(self):
        PTAH = ptah.get_settings(ptah.CFG_ID_PTAH, self.request.registry)
        return PTAH.get('sqlalchemy_initialized', False)

    @classmethod
    def addMetadata(cls, md, id, title=''):
        cls.metadata[id] = [md, title or id.capitalize()]

addMetadata = SQLAModule.addMetadata
addMetadata(ptah.get_base().metadata, 'psqla', 'Pyramid sqla')


class Table(object):

    def __init__(self, table, mod, request):
        self.__name__ = table.name
        self.__parent__ = mod

        self.table = table
        self.request = request

    def __getitem__(self, key):
        if key == 'add.html':
            raise KeyError(key)

        try:
            return Record(key, self.table, self, self.request)
        except:
            raise KeyError(key)


class Record(object):

    def __init__(self, pid, table, parent, request):
        self.pid = pid
        self.table = table
        self.request = request

        self.__name__ = str(pid)
        self.__parent__ = parent

        self.pname = None
        self.pcolumn = None
        for cl in table.columns:
            if cl.primary_key:
                self.pname = cl.name
                self.pcolumn = cl

        self.data = Session.query(table).filter(
            self.pcolumn == pid).one()


@view_config(
    context=SQLAModule, wrapper=ptah.wrap_layout(),
    renderer='ptah.manage:templates/sqla-index.pt')

class MainView(ptah.View):
    __doc__ = "sqlalchemy tables listing page."

    def printTable(self, table):
        columns = []
        for cl in table.columns:
            kwarg = []
            if cl.key != cl.name:
                kwarg.append('key') # pragma: no cover
            if cl.primary_key:
                kwarg.append('primary_key')
            if not cl.nullable:
                kwarg.append('nullable')
            if cl.default:
                kwarg.append('default')

            columns.append(
                "Column(%s)" % ', '.join(
                    [repr(cl.name)] + [repr(cl.type)] +
                    [repr(x) for x in cl.foreign_keys if x is not None] +
                    [repr(x) for x in cl.constraints] +
                    ["%s=%s" % (k, repr(getattr(cl, k))) for k in kwarg])
                )

        return ("Table(%s)" % repr(table.name), columns)

    def update(self):
        tables = []

        for id, (md, title) in self.context.metadata.items():
            data = []
            for name, table in md.tables.items():
                data.append((name, self.printTable(table)))

            tables.append((title, id, sorted(data)))

        self.tables = sorted(tables)


def get_inheritance(table):
    # inheritance
    inherits = []
    mapper = None
    for m, s in _mapper_registry.items():
        if m.local_table is table:
            mapper = m

    curr_mapper = mapper
    while curr_mapper is not None:
        curr_mapper = curr_mapper.inherits
        if curr_mapper is not None and \
                curr_mapper.local_table.name != table.name and \
                curr_mapper.local_table.name not in inherits:
            inherits.append(curr_mapper.local_table.name)

    inherits.reverse()
    return inherits


@view_config(
    context=Table, wrapper=ptah.wrap_layout(),
    renderer='ptah.manage:templates/sqla-table.pt')

class TableView(ptah.form.Form):
    __doc__ = "List table records."

    csrf = True
    page = ptah.Pagination(15)

    def update(self):
        table = self.table = self.context.table
        self.primary = None
        self.pcolumn = None
        self.uris = []

        self.inheritance = get_inheritance(table)

        if table.name == 'ptah_nodes' or self.inheritance:
            self.show_actions = False
        else:
            self.show_actions = True

        names = []
        for cl in table.columns:
            names.append(cl.name)
            if cl.primary_key:
                self.primary = cl.name
                self.pcolumn = cl
            if cl.info.get('uri'):
                self.uris.append(len(names)-1)

        res = super(TableView, self).update()

        request = self.request
        try:
            current = int(request.params.get('batch', None))
            if not current:
                current = 1

            request.session['table-current-batch'] = current
        except:
            current = request.session.get('table-current-batch')
            if not current:
                current = 1

        self.size = Session.query(table).count()
        self.current = current

        self.pages, self.prev, self.next = self.page(self.size, self.current)

        offset, limit = self.page.offset(current)
        self.data = Session.query(table).offset(offset).limit(limit).all()

        return res

    def quote(self, val):
        return url_quote_plus(val)

    def val(self, val):
        try:
            if isinstance(val, str): # pragma: no cover
                val = unicode(val, 'utf-8', 'ignore')
            elif not isinstance(val, unicode):
                val = str(val)
        except: # pragma: no cover
            val = "Can't show"
        return val[:100]

    @ptah.form.button('Add', actype=ptah.form.AC_PRIMARY)
    def add(self):
        return HTTPFound(location='add.html')

    @ptah.form.button('Remove', actype=ptah.form.AC_DANGER)
    def remove(self):
        self.validate_csrf_token()

        ids = []
        for id in self.request.POST.getall('rowid'):
            ids.append(id)

        if not ids:
            self.message('Please select records for removing.', 'warning')
            return

        self.table.delete(self.pcolumn.in_(ids)).execute()
        self.message('Select records have been removed.')


@view_config(
    context=Record, wrapper=ptah.wrap_layout(),
    renderer='ptah.manage:templates/sqla-edit.pt')

class EditRecord(ptah.form.Form):
    __doc__ = "Edit table record."

    csrf = True

    @reify
    def label(self):
        return 'record %s'%self.context.__name__

    @reify
    def fields(self):
        return ptah.build_sqla_fieldset(
            [(cl.name, cl) for cl in self.context.table.columns],
            skipPrimaryKey=True)

    def update(self):
        self.inheritance = get_inheritance(self.context.table)
        if self.context.table.name == 'ptah_nodes' or self.inheritance:
            self.show_remove = False
        else:
            self.show_remove = True

        return super(EditRecord, self).update()

    def form_content(self):
        data = {}
        for field in self.fields.fields():
            data[field.name] = getattr(
                self.context.data, field.name, field.default)
        return data

    @ptah.form.button('Cancel')
    def cancel_handler(self):
        return HTTPFound(location='..')

    @ptah.form.button('Modify', actype=ptah.form.AC_PRIMARY)
    def modify_handler(self):
        data, errors = self.extract()

        if errors:
            self.message(errors, 'form-error')
            return

        self.context.table.update(
            self.context.pcolumn == self.context.__name__, data).execute()
        self.message('Table record has been modified.', 'success')
        return HTTPFound(location='..')

    @ptah.form.button('Remove', actype=ptah.form.AC_DANGER,
                      condition=lambda form: form.show_remove)
    def remove_handler(self):
        self.validate_csrf_token()

        self.context.table.delete(
            self.context.pcolumn == self.context.__name__).execute()
        self.message('Table record has been removed.')
        return HTTPFound(location='..')



@view_config('add.html', context=Table, wrapper=ptah.wrap_layout())

class AddRecord(ptah.form.Form):
    """ Add new table record. """

    csrf = True

    @reify
    def label(self):
        return '%s: new record'%self.context.table.name

    @reify
    def fields(self):
        return ptah.build_sqla_fieldset(
            [(cl.name, cl) for cl in self.context.table.columns],
            skipPrimaryKey = True)

    @ptah.form.button('Create', actype=ptah.form.AC_PRIMARY)
    def create(self):
        data, errors = self.extract()

        if errors:
            self.message(errors, 'form-error')
            return

        try:
            self.context.table.insert(values = data).execute()
        except Exception as e: # pragma: no cover
            self.message(e, 'error')
            return

        self.message('Table record has been created.', 'success')
        return HTTPFound(location='.')

    @ptah.form.button('Save & Create new',
                      name='createmulti', actype=ptah.form.AC_PRIMARY)
    def create_multiple(self):
        data, errors = self.extract()

        if errors:
            self.message(errors, 'form-error')
            return

        try:
            self.context.table.insert(values = data).execute()
        except Exception as e: # pragma: no cover
            self.message(e, 'error')
            return

        self.message('Table record has been created.', 'success')

    @ptah.form.button('Back')
    def cancel(self):
        return HTTPFound(location='.')



import ptah
from ptah.testing import PtahTestCase
from pyramid.testing import DummyRequest


class TestUriView(PtahTestCase):

    _cleanup_mod = False

    def test_uri_view(self):
        from ptah.manage.uri import UriResolver

        request = DummyRequest(
            GET = {'uri': 'ptah-auth:superuser'})

        view = UriResolver(None, request)
        view.update()

        self.assertEqual(view.data[0]['name'],
                         'ptah.authentication.superuser_resolver')

    def test_uri_handler(self):
        from ptah.manage.uri import UriResolver

        request = DummyRequest(
            POST = {'form.buttons.show': 'Show'})

        view = UriResolver(None, request)
        view.update()

        request = DummyRequest(
            POST = {'form.buttons.show': 'Show', 'uri': 'ptah-auth:superuser'})

        view = UriResolver(None, request)
        view.update()

        self.assertEqual(view.data[0]['name'],
                         'ptah.authentication.superuser_resolver')



""" app management module """
import ptah
from ptah import view, form, cms
from pyramid.view import view_config
from pyramid.httpexceptions import HTTPFound
from pyramid.interfaces import IRequest, IRouteRequest

from ptah.manage import manage

MANAGE_APP_ROUTE = MANAGE_APP_CATEGORY = 'ptah-manage-app'


@manage.module('apps')
class ApplicationsModule(manage.PtahModule):
    __doc__ = 'A listing of all registered Ptah Applications.'

    title = 'Applications'

    def __getitem__(self, key):
        for id, factory in cms.get_app_factories().items():
            if factory.name == key:
                request = self.request
                request.request_iface = request.registry.getUtility(
                    IRouteRequest, name=MANAGE_APP_ROUTE)
                request.app_factory = factory
                app = factory()
                app.__parent__ = self
                app.__root_path__ = '/ptah-manage/apps/%s/'%app.__name__
                return app

        raise KeyError(key)

    def available(self):
        return bool(cms.get_app_factories())


@view_config(
    context=ApplicationsModule, wrapper=ptah.wrap_layout(),
    renderer='ptah.manage:templates/apps.pt')

class ApplicationsModuleView(ptah.View):
    """ Applications module default view """

    def update(self):
        factories = []
        for factory in cms.get_app_factories().values():
            factories.append((factory.title, factory))

        self.factories = [f for _t, f in sorted(factories)]


@ptah.layout(
    'ptah-manage', manage.PtahManageRoute,
    route_name=MANAGE_APP_ROUTE,
    renderer="ptah.manage:templates/ptah-manage.pt")

class AppLayout(manage.LayoutManage):
    """ Application module layout """


@ptah.layout(
    '', cms.Node, parent="ptah-manage",
    route_name=MANAGE_APP_ROUTE,
    renderer="ptah.manage:templates/apps-layout.pt")

class AppContentLayout(ptah.View):
    """ Application module content layout """

    def update(self):
        self.actions = ptah.list_uiactions(
            self.context, self.request, MANAGE_APP_CATEGORY)


@view_config(
    context=cms.Content,
    wrapper=ptah.wrap_layout(),
    route_name=MANAGE_APP_ROUTE,
    renderer="ptah.manage:templates/apps-contentview.pt")

class ViewForm(form.DisplayForm):

    @property
    def fields(self):
        return self.context.__type__.fieldset

    def form_content(self):
        data = {}
        for name, field in self.context.__type__.fieldset.items():
            data[name] = getattr(self.context, name, field.default)

        return data


@view_config(
    'sharing.html',
    context=cms.IContent,
    route_name=MANAGE_APP_ROUTE, wrapper=ptah.wrap_layout(),
    renderer='ptah.manage:templates/apps-sharing.pt')

class SharingForm(form.Form):
    """ Sharing form """

    csrf = True
    fields = form.Fieldset(
        form.FieldFactory(
            'text',
            'term',
            title = 'Search term',
            description = 'Searches users by login and email',
            missing = '',
            default = '')
        )


    users = None
    bsize = 15

    def form_content(self):
        return {'term': self.request.session.get('apps-sharing-term', '')}

    def get_principal(self, id):
        return ptah.resolve(id)

    def update(self):
        res = super(SharingForm, self).update()

        request = self.request
        context = self.context

        self.roles = [r for r in ptah.get_roles().values() if not r.system]
        self.local_roles = local_roles = context.__local_roles__

        term = request.session.get('apps-sharing-term', '')
        if term:
            self.users = list(ptah.search_principals(term))

        if 'form.buttons.save' in request.POST:
            users = []
            userdata = {}
            for attr, val in request.POST.items():
                if attr.startswith('user-'):
                    userId, roleId = attr[5:].rsplit('-',1)
                    data = userdata.setdefault(str(userId), [])
                    data.append(str(roleId))
                if attr.startswith('userid-'):
                    users.append(str(attr.split('userid-')[-1]))

            for uid in users:
                if userdata.get(uid):
                    local_roles[str(uid)] = userdata[uid]
                elif uid in local_roles:
                    del local_roles[uid]

            context.__local_roles__ = local_roles

        return res

    @form.button('Search', actype=form.AC_PRIMARY)
    def search(self):
        data, error = self.extract()

        if not data['term']:
            self.message('Please specify search term', 'warning')
            return

        self.request.session['apps-sharing-term'] = data['term']
        return HTTPFound(location = self.request.url)


ptah.uiaction(
    ptah.ILocalRolesAware, **{'id': 'view',
                              'title': 'View',
                              'action': '',
                              'category': MANAGE_APP_CATEGORY,
                              'sort_weight': 1.0})

ptah.uiaction(
    ptah.ILocalRolesAware, **{'id': 'sharing',
                              'title': 'Sharing',
                              'action': 'sharing.html',
                              'category': MANAGE_APP_CATEGORY,
                              'sort_weight': 10.0})



""" directives tests """
import sys
from pyramid import testing
from pyramid.exceptions import ConfigurationConflictError

from zope import interface
from zope.interface.registry import Components
from zope.interface.interfaces import IObjectEvent

import ptah
from ptah import config
from ptah.testing import TestCase


class BaseTesting(TestCase):

    def _init_ptah(self, *args, **kw):
        self.config.include('ptah')
        self.config.scan(self.__class__.__module__)
        self.config.commit()
        self.config.autocommit = True

    def setUp(self):
        self.config = testing.setUp(autocommit=False)
        self.config.include('ptah')
        self.registry = self.config.registry

    def tearDown(self):
        mod = sys.modules[self.__class__.__module__]
        if hasattr(mod, config.ATTACH_ATTR):
            delattr(mod, config.ATTACH_ATTR)

        testing.tearDown()


class IContext(interface.Interface):
    pass


class IContext2(interface.Interface):
    pass


class IAdapter(interface.Interface):
    pass


class Context(object):

    def __init__(self, iface):
        interface.directlyProvides(self, iface)


class TestAdaptsDirective(BaseTesting):

    def test_adapts_err(self):
        @ptah.adapter(IContext)
        @ptah.adapter(IContext)
        @interface.implementer(IAdapter)
        class TestClass(object):
            """ """

        err = None
        try:
            self._init_ptah()
        except Exception as e:
            err = e

        s = str(err)
        self.assertTrue(isinstance(err, ConfigurationConflictError))
        self.assertTrue('ptah.config:adapter' in s)
        self.assertTrue('test_directives' in s)

    def test_adapts(self):
        @ptah.adapter(IContext)
        @interface.implementer(IAdapter)
        class TestClass(object):
            def __init__(self, context):
                pass

        self._init_ptah()

        sm = self.registry
        adapters = sm.adapters.lookupAll((IContext,), IAdapter)

        self.assertTrue(len(adapters) == 1)
        self.assertTrue(adapters[0][0] == '')
        self.assertTrue(adapters[0][1] is TestClass)

        adapter = sm.getAdapter(Context(IContext), IAdapter)
        self.assertTrue(isinstance(adapter, TestClass))

    def test_adapts_named(self):
        @ptah.adapter(IContext, name='test')
        @interface.implementer(IAdapter)
        class TestClass(object):
            """ """

        self._init_ptah()

        sm = self.registry
        adapters = sm.adapters.lookupAll((IContext,), IAdapter)

        self.assertTrue(adapters[0][0] == 'test')
        self.assertTrue(adapters[0][1] is TestClass)

    def test_adapts_multiple(self):
        @ptah.adapter(IContext, IContext2)
        @interface.implementer(IAdapter)
        class TestClass(object):
            """ """
        self._init_ptah()

        sm = self.registry
        adapters = sm.adapters.lookupAll((IContext, IContext2), IAdapter)

        self.assertTrue(adapters[0][0] == '')
        self.assertTrue(adapters[0][1] is TestClass)

    def test_adapts_multiple_named(self):
        @ptah.adapter(IContext, IContext2, name='test')
        @interface.implementer(IAdapter)
        class TestClass(object):
            """ """
        self._init_ptah()

        sm = self.registry
        adapters = sm.adapters.lookupAll((IContext, IContext2), IAdapter)

        self.assertTrue(adapters[0][0] == 'test')
        self.assertTrue(adapters[0][1] is TestClass)

    def test_adapts_reinitialize(self):
        @ptah.adapter(IContext)
        @interface.implementer(IAdapter)
        class TestClass(object):
            """ """

        self._init_ptah()

        # reinstall
        sm = self.registry
        sm.__init__('base')

        adapters = sm.adapters.lookupAll((IContext,), IAdapter)
        self.assertTrue(len(adapters) == 0)

        self._init_ptah()

        sm = self.registry
        adapters = sm.adapters.lookupAll((IContext,), IAdapter)

        self.assertTrue(len(adapters) == 1)
        self.assertTrue(adapters[0][1] is TestClass)


class TestAdapterDirective(BaseTesting):

    def test_adapter(self):
        @config.adapter(IContext)
        @config.adapter(IContext2)
        @interface.implementer(IAdapter)
        def testAdapter(context):  # pragma: no cover
            pass

        self._init_ptah()

        sm = self.registry
        adapters = sm.adapters.lookupAll((IContext,), IAdapter)

        self.assertTrue(len(adapters) == 1)
        self.assertTrue(adapters[0][1] is testAdapter)

        adapters = sm.adapters.lookupAll((IContext2,), IAdapter)

        self.assertTrue(len(adapters) == 1)
        self.assertTrue(adapters[0][1] is testAdapter)

    def test_adapter_named(self):
        @config.adapter(IContext, name='test')
        @interface.implementer(IAdapter)
        def testAdapter(context):  # pragma: no cover
            pass

        self._init_ptah()

        sm = self.registry
        adapters = sm.adapters.lookupAll((IContext,), IAdapter)

        self.assertTrue(len(adapters) == 1)
        self.assertTrue(adapters[0][0] == 'test')
        self.assertTrue(adapters[0][1] is testAdapter)

    def test_adapter_multple(self):
        @config.adapter(IContext, IContext2, name='test')
        @interface.implementer(IAdapter)
        def testAdapter(context):  # pragma: no cover
            pass

        self._init_ptah()

        sm = self.registry
        adapters = sm.adapters.lookupAll((IContext, IContext2), IAdapter)

        self.assertTrue(len(adapters) == 1)
        self.assertTrue(adapters[0][0] == 'test')
        self.assertTrue(adapters[0][1] is testAdapter)

    def test_adapter_errs1(self):
        err = None
        try:
            @config.adapter(IContext, name='test')
            def testAdapter(context):  # pragma: no cover
                pass
        except Exception as e:
            err = e

        self.assertTrue(isinstance(err, TypeError))

    def test_adapter_errs2(self):
        @config.adapter(IContext)
        @config.adapter(IContext)
        @interface.implementer(IAdapter)
        def testAdapter(context):  # pragma: no cover
            pass

        self.assertRaises(ConfigurationConflictError, self._init_ptah)

    def test_adapter_reinitialize(self):
        @config.adapter(IContext)
        @config.adapter(IContext2)
        @interface.implementer(IAdapter)
        def testAdapter(context):  # pragma: no cover
            pass

        self._init_ptah()

        # reinstall
        sm = self.registry
        sm.__init__('base')

        self._init_ptah()

        adapters = sm.adapters.lookupAll((IContext,), IAdapter)

        self.assertTrue(len(adapters) == 1)
        self.assertTrue(adapters[0][1] is testAdapter)

        adapters = sm.adapters.lookupAll((IContext2,), IAdapter)

        self.assertTrue(len(adapters) == 1)
        self.assertTrue(adapters[0][1] is testAdapter)


class TestSubscriberDirective(BaseTesting):

    def test_subscriber(self):
        events = []

        @config.subscriber(IContext)
        def testHandler(*args):
            events.append(args)

        self._init_ptah()

        sm = self.registry
        sm.subscribers((Context(IContext),), None)

        self.assertTrue(len(events) == 1)

        sm.subscribers((Context(IContext2),), None)
        self.assertTrue(len(events) == 1)

    def test_subscriber_multple(self):
        events = []

        @config.subscriber(IContext)
        @config.subscriber(IContext2)
        def testHandler(*args):
            events.append(args)

        self._init_ptah()

        sm = self.registry
        sm.subscribers((Context(IContext),), None)
        sm.subscribers((Context(IContext2),), None)

        self.assertTrue(len(events) == 2)

    def test_subscriber_object(self):
        from zope.interface.interfaces import IObjectEvent, ObjectEvent

        events = []

        @config.subscriber(IContext, IObjectEvent)
        def testSubscriber(*args):
            events.append(args)

        self._init_ptah()

        sm = self.config.registry
        sm.subscribers((ObjectEvent(Context(IContext)),), None)

        self.assertTrue(len(events) == 1)
        self.assertTrue(len(events[0]) == 2)


class TestExtraDirective(BaseTesting):

    def test_action(self):
        info = config.DirectiveInfo()

        action = config.Action(None, discriminator=('test', ))

        self.assertIsNone(action.hash)

        info.attach(action)

        self.assertIsNotNone(action.hash)
        self.assertIsNotNone(hash(action))
        self.assertRaises(TypeError, info.attach, action)
        self.assertEqual(repr(action), '<Action "test">')



from ptah import config
from zope.interface.interfaces import ObjectEvent


class event(object):
    """ Register event object, it is used for introspection only. """

    ID_EVENT = 'ptah.config:event'

    #: Event name
    name = ''

    #: Event title
    title = ''

    #: Event category
    category = ''

    #: Event class or interface
    factory = None

    def __init__(self, title='', category=''):
        self.title = title
        self.category = category

        self.info = config.DirectiveInfo()

    def __call__(self, cls):
        self.factory = cls
        self.description = cls.__doc__
        self.name = '{0}.{1}'.format(cls.__module__, cls.__name__)

        discr = (self.ID_EVENT, self.name)
        intr = config.Introspectable(
            self.ID_EVENT, discr, self.description, self.title)
        intr['ev'] = self
        intr['name'] = self.name
        intr['codeinfo'] = self.info.codeinfo

        def _event(cfg, desc, intr):
            storage = cfg.get_cfg_storage(self.ID_EVENT)
            storage[desc.name] = desc
            storage[desc.factory] = desc

        self.info.attach(
            config.Action(_event, (self, intr),
                          discriminator=discr, introspectables=(intr,))
            )
        return cls


# settings related events

@event('Settings initializing event')
class SettingsInitializing(object):
    """ Settings initializing event """

    config = None
    registry = None

    def __init__(self, config, registry):
        self.config = config
        self.registry = registry


@event('Settings initialized event')
class SettingsInitialized(object):
    """ ptah sends this event when settings initialization is completed. """

    config = None
    registry = None

    def __init__(self, config, registry):
        self.config = config
        self.registry = registry


@event('Settings group modified event')
class SettingsGroupModified(object):
    """ ptah sends this event when settings group is modified. """

    def __init__(self, group):
        self.object = group


# uri events

@event('Uri invalidate event')
class UriInvalidateEvent(object):
    """ Uri object has been changed. """

    def __init__(self, uri):
        self.uri = uri


# principal events

class PrincipalEvent(object):
    """ base class for all principal related events """

    principal = None  # IPrincipal object

    def __init__(self, principal):  # pragma: no cover
        self.principal = principal


@event('Logged in event')
class LoggedInEvent(PrincipalEvent):
    """ User logged in to system."""


@event('Login failed event')
class LoginFailedEvent(PrincipalEvent):
    """ User login failed."""

    message = ''

    def __init__(self, principal, message=''):  # pragma: no cover
        self.principal = principal
        self.message = message


@event('Logged out event')
class LoggedOutEvent(PrincipalEvent):
    """ User logged out."""


@event('Reset password initiated event')
class ResetPasswordInitiatedEvent(PrincipalEvent):
    """ User has initiated password changeing."""


@event('User password has been changed')
class PrincipalPasswordChangedEvent(PrincipalEvent):
    """ User password has been changed. """


@event('Account validation event')
class PrincipalValidatedEvent(PrincipalEvent):
    """ Principal account has been validated."""


@event('Principal added event')
class PrincipalAddedEvent(PrincipalEvent):
    """ Principal added event """


@event('Principal registered event')
class PrincipalRegisteredEvent(PrincipalEvent):
    """ Principal registered event """


# content events

class ContentEvent(ObjectEvent):
    """ Base content event """

    object = None


@event('Content created event')
class ContentCreatedEvent(ContentEvent):
    """ :py:class:`ptah.cms.TypeInformation` will send event during create().
    """


@event('Content added event')
class ContentAddedEvent(ContentEvent):
    """ :py:class:`ptah.cms.Container` will send event when content has been
        created through containers __setitem__ method.
    """


@event('Content moved event')
class ContentMovedEvent(ContentEvent):
    """ :py:class:`ptah.cms.Container` will send event when content has moved.
    """


@event('Content modified event')
class ContentModifiedEvent(ContentEvent):
    """ :py:class:`ptah.cms.Content` will send event during update().
    """


@event('Content deleting event')
class ContentDeletingEvent(ContentEvent):
    """ :py:class:`ptah.cms.Container` will send event before content has been
        deleted through containers __delitem__ method.
    """


# db schema creation
class BeforeCreateDbSchema(object):
    """ :py:data:`ptah.POPULATE_DB_SCHEMA` populate step sends event before
    tables have been created.

    ``registry``: Pyramid registry object
    """

    def __init__(self, registry):
        self.registry = registry



from datetime import datetime
from ptah.testing import TestCase


class TestJson(TestCase):

    def test_datetime(self):
        from ptah import util
        dt = datetime(2011, 10, 1, 1, 56)

        self.assertEqual(util.dthandler(dt), '2011-10-01T01:56:00')
        self.assertIsNone(util.dthandler(object()))

    def test_json(self):
        from ptah import util

        data = {'date': datetime(2011, 10, 1, 1, 56),
                'int': 10,
                'str': 'string'}

        self.assertEqual(
            util.json.dumps(data),
            '{"date":"2011-10-01T01:56:00","int":10,"str":"string"}')

        self.assertEqual(
            util.json.loads('{"int":10,"str":"string"}'),
            {'int': 10, 'str': 'string'})



""" base view class with access to various api's """
import cgi
import logging
from zope.interface import Interface
from pyramid.decorator import reify
from pyramid.compat import string_types
from pyramid.renderers import RendererHelper
from pyramid.config.views import DefaultViewMapper

import ptah.view
from ptah import config
from ptah.formatter import format
from ptah.library import include, render_includes

log = logging.getLogger('ptah.view')

SNIPPET_ID = 'ptah.view:snippet'


class View(object):
    """ Base view """

    __name__ = ''
    __parent__ = None

    format = format

    def __init__(self, context, request):
        self.context = context
        self.request = request
        self.__parent__ = context

    @reify
    def application_url(self):
        url = self.request.application_url
        if url.endswith('/'):
            url = url[:-1]
        return url

    def update(self):
        return {}

    def __call__(self):
        result = self.update()
        if result is None:
            result = {}

        return result

    def include(self, name):
        include(self.request, name)

    def render_includes(self):
        return render_includes(self.request)

    def message(self, msg, type='info'):
        ptah.view.add_message(self.request, msg, type)

    def render_messages(self):
        return ptah.view.render_messages(self.request)

    def snippet(self, stype, context=None):
        if context is None:
            context = self.context

        request = self.request

        try:
            snippet = request.registry.queryMultiAdapter(
                (context, request), ISnippet, stype)

            if snippet is not None:
                return snippet
        except Exception as e:
            log.exception(str(e))

        return ''


class ISnippet(Interface):
    """ marker interface """


def render_snippet(name, context, request):
    """Render snippet

    :param name: Name of snippet
    :param context: Context for snippet
    :param request: Current request
    :rtype: string with rendered html

    """
    try:
        return request.registry.getMultiAdapter(
            (context, request), ISnippet, name)
    except Exception as e:
        log.exception(str(e))
        raise


class snippet(object):
    """Register snippet.

    :param name: Snippet name
    :param context: Snippet context
    :param renderer: Pyramid renderer

    Example:

    .. code-block:: python

      import ptah

      @ptah.snippet('test', Context, renderer='...:test.pt')
      def snippet(request):
          ...

      ptah.render_snippet('test', Context(), request)

    To render snippet use :py:func:`ptah.render_snippet` or
    :py:class:`ptah.View.snippet`.

    """

    def __init__(self, name, context=None, renderer=None, __depth=1):
        self.info = config.DirectiveInfo(__depth)

        self.discr = discr = (SNIPPET_ID, name, context)

        intr = config.Introspectable(SNIPPET_ID, discr, name, SNIPPET_ID)
        intr['name'] = name
        intr['context'] = context
        intr['renderer'] = renderer
        intr['module'] = self.info.module.__name__
        intr['codeinfo'] = self.info.codeinfo
        self.intr = intr

    @classmethod
    def register(cls, name, context=None, view=None, renderer=None):
        """ Register snippet.

        :param name: Snippet name
        :param context: Snippet context
        :param view: View implementation
        :param renderer: Pyramid renderer

        .. code-block:: python

          def snippet(request):
              ...

          ptah.snippet.register('test', Context, view=snippet):
              ...

        """
        return snippet(name, context, renderer, 2)(view)

    @classmethod
    def pyramid(cls, cfg, name, context=None, view=None, renderer=None):
        """ Pyramid `ptah_snippet` directive.

        :param name: Snippet name
        :param context: Snippet context
        :param view: View implementation
        :param renderer: Pyramid renderer

        .. code-block:: python

          config = Configurator()
          config.include('ptah')

          config.ptah_snippet('test', view=snippet, renderer='.../test.pt')
        """
        return snippet(name, context, renderer, 3)(view, cfg)

    def _register(self, cfg):
        intr = self.intr

        renderer = intr['renderer']
        if isinstance(renderer, string_types):
            renderer = RendererHelper(name=renderer, registry=cfg.registry)

        view = intr['view']
        if view is None:
            view = ptah.View

        # register snippet
        context = intr['context']
        if context is None:
            context = Interface

        cfg.registry.registerAdapter(
            SnippetRenderer(view, context, renderer),
            [context, Interface], ISnippet, name=intr['name'])

    def __call__(self, view, cfg=None):
        intr = self.intr
        intr['view'] = view

        self.info.attach(
            config.Action(
                self._register,
                discriminator=self.discr, introspectables=(intr,)),
            cfg)
        return view


class SnippetRenderer(object):

    def __init__(self, view, context, renderer):
        self.view = view
        self.context = context
        self.renderer = renderer

        mapper = getattr(view, '__view_mapper__', DefaultViewMapper)
        self.mapped_view = mapper()(view)

    def __call__(self, context, request):
        orig_view = getattr(request, '__view__', None)
        value = self.mapped_view(context, request)
        if self.renderer is None:
            request.__view__ = orig_view
            return value

        if value is None:
            value = {}

        system = {'view': getattr(request, '__view__', None),
                  'renderer_info': self.renderer,
                  'context': context,
                  'request': request}

        result = self.renderer.render(value, system, request)
        request.__view__ = orig_view
        return result


def add_message(request, msg, type='info'):
    """ Add status message

    Predefined message types

    * info

    * success

    * warning

    * error

    """
    message = Message(msg, request)
    try:
        msg = request.registry.getMultiAdapter(
            (message, request), ISnippet, type)
    except Exception as e:
        log = logging.getLogger('ptah.view')
        log.exception(str(e))
    request.session.flash(msg, 'status')


def render_message(request, msg, type='info'):
    """ Render message, return html """
    message = Message(msg, request)
    try:
        msg = request.registry.getMultiAdapter(
            (message, request), ISnippet, type)
    except Exception as e:
        log = logging.getLogger('ptah.view')
        log.exception(str(e))

    return msg


def render_messages(request):
    """ Render previously added messages """
    return ''.join(request.session.pop_flash('status'))


class Message(object):
    """ Message context """

    def __init__(self, message, request):
        self.message = message
        self.request = request


snippet.register(
    'info', Message, renderer='ptah.view:templates/msg-info.pt')

snippet.register(
    'success', Message, renderer='ptah.view:templates/msg-success.pt')

snippet.register(
    'warning', Message, renderer='ptah.view:templates/msg-warning.pt')


@snippet('error', Message, renderer='ptah.view:templates/msg-error.pt')
def errorMessage(context, request):
    e = context.message

    if isinstance(e, Exception):
        message = '%s: %s'%(
            e.__class__.__name__, cgi.escape(str(e), True))
    else:
        message = e

    return {'message': message}



""" Node implementation """
import sqlalchemy as sa
from collections import OrderedDict
from pyramid.compat import text_type
from zope.interface import implementer

import ptah
from ptah.cms import action
from ptah.cms.permissions import View
from ptah.cms.interfaces import NotFound, Forbidden
from ptah.cms.interfaces import INode, IApplicationPolicy


@implementer(INode,
             ptah.IACLsAware,
             ptah.IOwnersAware,
             ptah.ILocalRolesAware)

class Node(ptah.get_base()):
    """ Base class for persistent objects.

    .. attribute:: __uri__

       Unique object id. **Required**

    .. attribute:: __type__

       Type information object :py:class:`ptah.cms.TypeInformation`

    .. attribute:: __parent__

       Parent of node. Ptah doesn't load `__parent__` automatically.
       To load node parents use :py:func:`ptah.cms.load_parents` function.

    .. attribute:: __owner__

       URI of owner principal. It possible to load principal object
       by using :py:func:`ptah.resolve` function.

    .. attribute:: __local_roles__

       :py:class:`ptah.JsonDictType` which contains a principal uri as a
       key and a sequence of role's granted to principal for Node.

    .. attribute:: __acls__

       a :py:class:`ptah.JsonListType` of :py:class:`ptah.ACL` strings
       registered with security machinery.

    .. attribute:: __annotations__

       a dictionary which contains strings arbitrary annotations data.

    .. attribute:: __uri_factory__

       function which will return value for __uri__.  the uri must be
       resolvable by using :py:func:`ptah.resolve` function.

    """

    __tablename__ = 'ptah_nodes'

    __id__ = sa.Column('id', sa.Integer, primary_key=True)
    __type_id__ = sa.Column('type', sa.String(128), info={'uri':True})
    __type__ = None

    __uri__ = sa.Column('uri', sa.String(255), unique=True,
                        nullable=False, info={'uri':True})
    __parent_uri__ = sa.Column('parent', sa.String(255),
                               sa.ForeignKey(__uri__), info={'uri': True})

    __owner__ = sa.Column('owner', sa.String(255), default='',info={'uri':True})
    __local_roles__ = sa.Column('roles', ptah.JsonDictType(), default={})
    __acls__ = sa.Column('acls', ptah.JsonListType(), default=[])
    __annotations__ = sa.Column('annotations', ptah.JsonDictType(),default={})

    __children__ = sa.orm.relationship(
        'Node', backref=sa.orm.backref('__parent_ref__', remote_side=[__uri__]))

    __mapper_args__ = {'polymorphic_on': __type_id__}

    __parent__ = None
    __uri_factory__ = None

    __acl__ = ptah.ACLsProperty()

    def __init__(self, **kw):
        self.__owners__ = []
        self.__local_roles__ = {}
        self.__permissions__ = []

        for attr, value in kw.items():
            setattr(self, attr, value)

        if '__parent__' in kw and kw['__parent__'] is not None:
            self.__parent_uri__ = kw['__parent__'].__uri__

        try:
            self.__uri__ = self.__uri_factory__()
        except TypeError:
            raise TypeError(
                'Subclass of Node has to override __uri_factory__')

    @action(permission=View)
    def info(self):
        info = OrderedDict(
            (('__type__', self.__type_id__),
             ('__content__', False),
             ('__uri__', self.__uri__),
             ('__parents__', [p.__uri__ for p in load_parents(self)]),
             ))

        return info

    def get_session(self):
        return sa.orm.object_session(self)


def load(uri, permission=None):
    """ Load node by `uri` and initialize __parent__ attributes. Also checks
    permission if permissin is specified.

    :param uri: Node uri
    :param permission: Check permission on node object
    :type permission: Permission id or None
    :raise KeyError: Node with this uri is not found.
    :raise Forbidden: If current principal doesn't pass permission check on loaded node.
    """
    item = ptah.resolve(uri)

    if item is not None:
        load_parents(item)

        if permission is not None:
            if not ptah.check_permission(permission, item):
                raise Forbidden()
    else:
        raise NotFound(uri)

    return item


def load_parents(node):
    """ Load and initialize `__parent__` attribute for node.
    Returns list of loaded parents.

    :param node: ptah.cms.Node node
    """
    parents = []
    parent = node
    while parent is not None:
        if not isinstance(parent, Node):
            break

        if parent.__parent__ is None:
            parent.__parent__ = parent.__parent_ref__

        parent = parent.__parent__
        if parent is not None:
            parents.append(parent)

    root = node if not parents else parents[-1]
    if not IApplicationPolicy.providedBy(root):
        try:
            root.__parent__ = get_policy()
        except AttributeError:
            # __parent__ is read onl
            pass

    return [p for p in parents if isinstance(p, Node)]


KEY = '__ptahcms_policy__'

def get_policy():
    """ Get current policy """
    return ptah.tldata.get(KEY)

def set_policy(policy):
    """ Set current policy """
    return ptah.tldata.set(KEY, policy)
